(***********************************************************************
  Author:	Robert Rossmair
		Wagelwitz, Rodaer Str. 6
		04688 Mutzschen
		Germany

	e-mail:	Robert.Rossmair@t-online.de
	http://home.t-online.de/home/Robert.Rossmair/

  Module:	ZoomDisp

  Version:	3.2.5, 18-May-1999

  Compatibility: Delphi 2.0, Delphi 3 and Delphi 4

  Description:  Implements T(Custom)ZoomDisplay and related stuff

  History:
    18-May-99: Version 3.2.5
        Some fixing.
        TCustomZoomDisplay:
        Method CheckFileImageVersion destilled from Loaded
    16-Mar-99: Version 3.2.4
	overrides TMarker.ValidateContainer instead
	of TMarker.SetParent for Delphi version >= 3
    21-Feb-99: Version 3.2.3
	SetWorld made virtual
    13-Feb-99: Version 3.2.2
	Fixed bug in UpdateWorldWindow by replacing
    ------8<------8<------8<------8<------8<------8<------8<------
    if not EqualRect(FPrevWindow, FWindow) then
    begin
      WindowChange;
      FPrevWindow := FWindow;
    end;
    ------8<------8<------8<------8<------8<------8<------8<------
    by
    ------8<------8<------8<------8<------8<------8<------8<------
    if not EqualRect(FPrevWindow, FWindow) then
    begin
      FPrevWindow := FWindow;
      WindowChange;
    end;
    ------8<------8<------8<------8<------8<------8<------8<------
    22-Aug-98: Version 3.2.1
      License changed to GNU Library General Public License to
      facilitate the use in commercial programs.
      Changed declaration of T16BitSampleArray and T24BitSampleArray
      for the sake of Delphi 4 compatibility.
    Version 3.2.0:
      Changed parameter list of OnWorldPosChange event
      Changed evaluation of SelectionColor so that it reflects the
      visible background color of the selected range.
      New Info property.
      TSignalDisplayEditor is a component editor for
      TCustomZoomDisplay which overrides the Edit method to change
      the double-click behavior in design mode. Now a double-click
      brings up a component information dialog (version and
      copyright information).
      Locking WaveformCanvas in TCustomZoomDisplay.Paint prevents
      the device context's viewport origin from being changed by
      e.g. Graphics.FreeMemoryContexts.
      New TDisplayScale.ShowGrid property.
      Scales now visible in design mode.
      Some fixes.
    Version 3.1.0:
      Possibility to change the zoom mode temporarily by pressing
      the Shift key
      TCustomZoomDisplay: Fixed BeginZoom, SetColor
      Fixed TDisplayScale
    Version 3: Scales!
      New types:
      * TDisplayScale
      * TGetScalePointProc
      Changes in T(Custom)ZoomDisplay:
      New properties:
      * HorzScale
      * VertScale
      * MinVertScaleWidth
      * Viewport
      * OnGetHorzScalePoint
      * OnGetVertScalePoint
      Optimizations in PlotByteArray, PlotSmallintArray
      Buffered and MonochromeBitmap now True by default
    30-Apr-98: Version 2.1.0
      Changes in TCustomZoomDisplay:
      * New property PaintMode
      * TAmpFunc result type changed; consequential changes in
        PlotSamples
      * removed property WorldPoint (superfluous)
      * renamed Canvas to WaveformCanvas
    17-Apr-98: Version 2.0.6: TCustomZoomDisplay.SetWorld enhanced
        (EqualsW?, EqualsMaxW? stuff)
    16-Apr-98: Version 2.0.5: Bug fix in PlotSamples method
    13-Apr-98: Version 2.0.4:
        TMarker:
          New: SetKind; To uncouple the Visible property from
          side effects of its internal use, methods SetVisible,
          SetInheritedVisible have been added; this required
          changes in T{Horz|Vert}Marker.UpdatePos;
        TCustomZoomDisplay:
          Changed BoundWindow so that FXPos/FYPos are updated if necessary
          Changed SetWindow so that FXPos is updated if necessary
          Fixed DragOver, CMMouseLeave
    10-Apr-98: Version 2.0.3:
	Changed WorldX/Y result type to TWorldCoord
        Removed TZoomDisplay.OnEndDrag
        Added TMarker.OnEndDrag
    07-Apr-98: Version 2.0.2:
        Fixed THorz/VertMarker.UpdatePos so that they take into account
        changes in Height resp. Width of the marker's Parent...
    07-Apr-98: Version 2.0.0:
	Some measures have been taken to make the off-screen bitmap
	refresh only when necessary.
	This requires a means to trigger an off-screen bitmap update
	manually in situations where TCustomZoomDisplay cannot
	recognize that the bitmap needs to be updated; that's the task
	of TCustomZoomDisplay.InvalidateWorldWindow.
	In the course of revising line-drawing routines DeviceLineTo/
	DeviceMoveTo have been removed, the result type of DeviceX/Y
	changed to Double.
	Removed OnChange event; use OnWindowChange instead.
	Minor bug fixes.
      New features of TCustomZoomDisplay:
	* properties SelectionStart, SelectionEnd, OnSelectionChange,
	  SelectionColor, DrawSelectionMode, OnMouseLeave
	* methods SetSelection, Plot24BitSamples, PlotSamples
    21-Mar-98:	Version 1.4.1
	changed visibility of DeviceX/Y methods from protected to public;
	fixed bug in T{Horz|Vert}Marker.SetAlign (which used to align
	to Parent.BoundsRect instead of Parent.ClientRect before)
    18-Mar-98:	Version 1.4.0
	TCustomZoomDisplay: replaced property SaveImageAsWMF
	by ImageFileFormat;
	minor bug fix: SetBuffered used to create a bitmap
	too small in size (by 1 Pixel in height and width).
    18-Mar-98:	Version 1.3.3
	Implemented OnWindowChange event.
	Bug fixes:
	fixed TCustomZoomDisplay.GetBitmap in respect to MonochromeBitmap.
	In some cases (e.g. PenChanged) I referenced FBitmap without
	ensuring FBitmap <> nil; fixed that.
	TCustomZoomDisplay.EndZoom: BeginZoom sets Brush.Style to
	bsClear, which has the side effect to change Brush.Color to
	clWhite; hence I replaced "Brush.Style := bsSolid" by
	"Brush.Color := Color" with the side effect of changing
	Brush.Style to bsSolid.
    17-Mar-98:	Version 1.3.2
	fixed TCustomZoomDisplay.Paint in respect to MonochromeBitmap
    13-Mar-98:	Version 1.3.1
	removed some superfluous stuff
    12-Mar-98:	Version 1.3.0
	introduces MonochromeBitmap
    11-Mar-98:	Version 1.2.1
	removed TCustomZoomDisplay.SetMarkerColorOf(),
	TCustomZoomDisplay.SetMarkerHintOf();
	IterateMarkers may be used instead
    10-Mar-98:	Version 1.2.0
	PlotByteArray implemented. Some details changed.
    27-Nov-97:	Version 1.1.1
	changed type to TWorldCoord for:
	TMarker.Position,
	TCustomZoomDisplay.DeviceX/Y arguments,
	TCustomZoomDisplay.MoveTo/LineTo arguments,
    26-Nov-97: Version 1.1.0
	added TCustomZoomDisplay.Error,
	TCustomZoomDisplay.PlotSmallintArray,
	TCustomZoomDisplay.PlotIntArray
    21-Nov-97: Version 1.0.12
	TCustomZoomDisplay.IterateMarkers
    20-Nov-97: Version 1.0.11
	added THorzMarker, TVertMarker classes
    17-Nov-97: Version 1.0.10
	added TMarker.Locked property
	added TCustomZoomDisplay.SetMarkerColorOf(),
    15-Nov-97: Version 1.0.9
	added TCustomZoomDisplay.SetMarkerHintOf(),
	      TCustomZoomDisplay.UpdateMarkerPosOf()
	      TMarker.SetOnGetposition()
    14-Nov-97: Version 1.0.8
	added TCustomZoomDisplay.DeleteMarkersMarking()
    10-Nov-97: Version 1.0.7
	added TMarker.Marked property
	published all Events inherited from TControl
    09-Nov-97: Version 1.0.6
	added TCustomZoomDisplay.FDoDrawMarkerPositions
    08-Nov-97: Version 1.0.5
	moved resource initialization (cursors)
	into initialization section
    01-Nov-97: Version 1.0.4
	added FRestoreColors and related stuff to equalize monochrome
	bitmap palette problems in Delphi 3.
    29-Oct-97: Version 1.0.3
	added property TCustomZoomDisplay.Bitmap to make the paint
	buffer visible to the public.
    27-Oct-97: Version 1.0.2
	added condition "if FPosition <> Value then"
	to TMarker.SetPosition
    10-Oct-97: Version 1.0.1
	added condition "if not (csDestroying in ComponentState)" to
	TCustomZoomDisplay.WMSize. Without this condition an
	EAccessViolation is raised on program exit if:
	1. Two MDI child windows with TZoomDisplay controls placed on
	   each of them are open,
	2. one of these windows is maximized.
    17-Sep-97: Version 1.0.0

  Copyright  1993-1999  Robert Rossmair

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License (file COPYING.Lib) along with this library; if not, write
  to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
  MA 02139, USA.
 ***********************************************************************)
unit ZoomDisp;

interface
{ $DEFINE Profile}
uses
  Windows, Messages, SysUtils, Classes, Graphics,
  Controls, Forms, ExtCtrls, RRMath, RRColors,
  CompInfo, CopyrInf,
  RRScales;

const
  crBase	= 10260;
  crCrossHairs	= crBase;
  crHand        = crBase+1;
  crGrip        = crBase+2;
  crZoomIn	= crBase+3;
  crZoomOut	= crBase+4;

  High24BitSample = 1 shl 23-1;
  Low24BitSample = -1 shl 23;
type
  TWorldCoord = Double;
  TWorldPoint = record
    X, Y: TWorldCoord;
  end;
  TDoublePoint = TWorldPoint;
  T24BitSample = packed record
    case Integer of
    0: (Byte1, Byte2, Byte3: Byte);
    1: (LoWord: Word; HiByte: Shortint);
    2: (LoByte: Byte; HiWord: Smallint);
  end;
  T16BitSampleArray = array[0..(High(Cardinal)div 2) div SizeOf(Smallint)-1] of Smallint;
  T24BitSampleArray = array[0..(High(Cardinal)div 2) div SizeOf(T24BitSample)-1] of T24BitSample;
  P16BitSampleArray = ^T16BitSampleArray;
  P24BitSampleArray = ^T24BitSampleArray;

  EWindowSize = class(Exception);
  EZoomDisplayError = class(Exception);
  TCustomZoomDisplay = class;

  TPageFraction = 1..High(Smallint);
  TDisplayScrollBar = class(TPersistent)
  private
    FDisplay: TCustomZoomDisplay;
    FPageFraction: TPageFraction;
    FKind: TScrollBarKind;
    FVisible: Boolean;
    FPersistent: Boolean;
    FTracking: Boolean;
    constructor Create(ADisplay: TCustomZoomDisplay; AKind: TScrollBarKind);
    function GetMin: Integer;
    function GetMax: Integer;
    function GetMask(Flags: UInt): UInt;
    function GetPageSize: Integer;
    function GetPosition: Integer;
    function GetScrollPos: Integer;
    procedure SetPersistent(Value: Boolean);
    procedure SetPosition(Value: Integer);
    procedure SetVisible(Value: Boolean);
    procedure Update;
  public
    procedure Assign(Source: TPersistent); override;
    property Kind: TScrollBarKind read FKind;
    property Min: Integer read GetMin;
    property Max: Integer read GetMax;
    property PageSize: Integer read GetPageSize;
  published
    property PageFraction: TPageFraction read FPageFraction write FPageFraction default 10;
    property Persistent: Boolean read FPersistent write SetPersistent;
    property Position: Integer read GetPosition write SetPosition default 0;
    property Tracking: Boolean read FTracking write FTracking default False;
    property Visible: Boolean read FVisible write SetVisible default True;
  end;

  TGetScalePointProc = procedure(Sender: TObject; var Value: Extended) of object;

  TDisplayScale = class(TScale)
  private
    FDisplay: TCustomZoomDisplay;
    FShowGrid: Boolean;
    procedure SetShowGrid(Value: Boolean);
  protected
    constructor Create(ADisplay: TCustomZoomDisplay);
    procedure Changed; override;
    property StartPointX;
    property StartPointY;
    property EndPointX;
    property EndPointY;
  published
    property Font;
    property MinScaleLineDistance;
    property ShowAxis;
    property ShowGrid: Boolean read FShowGrid write SetShowGrid;
    property Visible;
  end;

  TMarker = class;
  TMarkerProc = procedure(Marker: TMarker; Param: DWord);

  TClipCodes = (ccCenter, ccLeft, ccRight, ccTop, ccBottom);
  TClipCode = set of TClipCodes;
  TZoomMode = (zmZoomOff, zmZoomIn, zmZoomOut);
  TZoomStartMode = (zmAutomatic, zmManual);
  TPosEvent = procedure(Sender: TObject; X, Y: TWorldCoord) of object;
  TAmpFunc = function(var P: Pointer): TWorldCoord of object;
  TImageFileFormat = (imBMP, imEMF, imWMF);
  TDrawSelectionMode = (smNop, smInvert, smPatInvert);
  TSignalPaintMode = (spHollow, spSolid);

  TCustomZoomDisplay = class(TCustomControl)
  private
    FWorld: TRect;
    FWindow: TRect;
    FPrevWindow: TRect;
    FHorzScrollBar: TDisplayScrollBar;
    FVertScrollBar: TDisplayScrollBar;
    FXIncrement: Cardinal;
    FYIncrement: Cardinal;
    FMinWindowHeight: Integer;
    FMinWindowWidth: Integer;
    FMaxWindowHeight: Integer;
    FMaxWindowWidth: Integer;
    FSelectionStart: Integer;
    FSelectionEnd: Integer;
    FXPos: Integer;
    FYPos: Integer;
    FMouseX: Integer;
    FMouseY: Integer;
    FViewport: record
      X, Y: Integer;
      H, dH: Integer;
      W, dW: Integer;
    end;
    FPen: TPen;
    FBitmap: TBitmap;
    FSelBrush: TBrush;
    FGridPen: TPen;
    FHorzScale: TDisplayScale;
    FVertScale: TDisplayScale;
    FMinVertScaleWidth: Integer;
    FHLSample: Integer;
    FHighlightedX: Integer;
    FHighlightColor: TColor;
    FSelectionColor: TColor;
    FDecentContrast: TColor;
    FMetafileCanvas: TMetafileCanvas;
    FOnMouseLeave: TNotifyEvent;
    FOnPaint: TNotifyEvent;
    FOnSelectionChange: TNotifyEvent;
    FOnWindowChange: TNotifyEvent;
    FOnWorldPosChange: TPosEvent;
    FOnZoomMin: TNotifyEvent;
    FOnZoomMax: TNotifyEvent;
    FOnGetHorzScalePoint: TGetScalePointProc;
    FOnGetVertScalePoint: TGetScalePointProc;
    FCurrentPos: TDoublePoint;
    FDragPos: TPoint;
    FInfo: TComponentInfo;
    FCurrentCode: TClipCode;
    FAutoScrollHorz: Boolean;
    FAutoScrollVert: Boolean;
    FZoomMode: TZoomMode;
    FZoomStartMode: TZoomStartMode;
    FZooming: Boolean;
    FBitmapInvalid: Boolean;
    FWindowInvalid: Boolean;
    FRestoreColors: Boolean;
    FMouseOver: Boolean;
    FDoDrawMarkerPositions: Boolean;
    FInWaveformPaint: Boolean;
    FMonochromeBitmap: Boolean;
    FImageFileFormat: TImageFileFormat;
    FDrawSelectionMode: TDrawSelectionMode;
    FPaintMode: TSignalPaintMode;
    FDragLineVisible: Boolean;
    FHLSampleValid: Boolean;
    FAlternateZoomModes: Boolean;
    // Result := <AWindow changed>
    function BoundWindow(var AWindow: TRect): Boolean;
    function DevicePositionCode(P: TDoublePoint): TClipCode;
    procedure DoAutoScroll;
    procedure DoLineTo(DC: hDC; X, Y: TWorldCoord);
    procedure DoMouseLeave;
    procedure DrawHLSample;
    function HighlightedRect: TRect;
    procedure InvalidateHLSample;
    procedure SetHighlightedSample(Value: Integer);
    procedure UpdateHighlightedX;
    procedure HorzDragLine(X, Y: Integer; DragLineVisible: Boolean);
    procedure VertDragLine(X, Y: Integer; DragLineVisible: Boolean);
    procedure DrawMarkerPositions;
    procedure DrawScales;
    procedure DrawSelection;
    procedure Error(Text: string);
    function GetBitmap: TBitmap;
    function GetBuffered: Boolean;
    function GetSelectionRect: TRect;
    function GetViewport: TRect;
    function GetWaveformCanvas: TCanvas;
    function GetWindowWidth: Integer;
    function GetWindowHeight: Integer;
    function GetWorldWidth: Integer;
    function GetWorldHeight: Integer;
    procedure GridPenChanged(Sender: TObject);
    procedure MoveWindow(X, Y: Integer);
    procedure PenChanged(Sender: TObject);
    // function PositionCode(P: TPoint): TClipCode;
    procedure HScroll(var Message: TWMScroll);
    procedure VScroll(var Message: TWMScroll);
    procedure ScaleDrawGrid(Sender: TObject; ACanvas: TCanvas; Points: array of TPoint);
    procedure ScrollTo(X, Y: Integer);
    procedure SetBuffered(Value: Boolean);
    procedure SetGridPen(Value: TPen);
    procedure SetHighlightColor(Value: TColor);
    procedure SetInfo(Value: TComponentInfo);
    procedure SetPaintMode(Value: TSignalPaintMode);
    procedure SetPen(Value: TPen);
    procedure SetMinVertScaleWidth(Value: Integer);
    procedure SetMinWindowWidth(Value: Integer);
    procedure SetMinWindowHeight(Value: Integer);
    procedure SetMaxWindowWidth(Value: Integer);
    procedure SetMaxWindowHeight(Value: Integer);
    procedure SetMonochromeBitmap(Value: Boolean);
    procedure SetHorzScale(Value: TDisplayScale);
    procedure SetVertScale(Value: TDisplayScale);
    procedure SetHorzScrollBar(Value: TDisplayScrollBar);
    procedure SetVertScrollBar(Value: TDisplayScrollBar);
    procedure SetSelectionStart(Value: Integer);
    procedure SetSelectionEnd(Value: Integer);
    procedure SetDrawSelectionMode(Value: TDrawSelectionMode);
    procedure SetSelectionColor(Value: TColor);
    procedure SetViewport(Left, Top, Width, Height: Integer);
    procedure SetWindow(Value: TRect);
    procedure SetWindowLeft(Value: Integer);
    procedure SetWindowWidth(Value: Integer);
    procedure SetWindowTop(Value: Integer);
    procedure SetWindowHeight(Value: Integer);
    procedure SetWorldLeft(Value: Integer);
    procedure SetWorldWidth(Value: Integer);
    procedure SetWorldTop(Value: Integer);
    procedure SetWorldHeight(Value: Integer);
    procedure UpdateScrollBars;
    procedure ContinueZoom(X, Y: Integer);
    procedure EndZoom;
    procedure DrawZoomBox;
    procedure UpdateMarkers;
    procedure UpdateHorzScale;
    procedure UpdateVertScale;
    procedure UpdateScales;
    procedure UpdateSelectionBrush;
    procedure UpdateViewport;
    procedure VerticalLine(DC: hDC; X, Ymin, Ymax: Integer);
    procedure WindowChange;
    procedure WorldPosChanged(MouseX, MouseY: Integer);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMDrag(var Message: TCMDrag); message CM_DRAG;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure WMNCMouseMove(var Message: TMessage); message WM_NCMOUSEMOVE;
    procedure WMSize(var Message: TWMSize); message wm_Size;
    procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
  protected
    destructor Destroy; override;
    class function Version: Integer; dynamic;
    procedure CheckFileImageVersion; dynamic;
    procedure CreateWnd; override;
    procedure DoWaveformPaint; dynamic;
    procedure DragDrop(Source: TObject; X, Y: Integer); override;
    procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean); override;
    function GetPalette: HPalette; override;
    procedure GetZoomWindow(var AWindow: TRect);
    procedure Loaded; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);	override;
    procedure Paint; override;
    procedure RealizeZoom; virtual;
    procedure ScrollBy(Dx, Dy: Integer);
    procedure SetWorld(const Value: TRect); virtual;
    procedure UpdateIncrements; virtual;
    function UpdateWorldWindow: boolean; virtual;
    procedure WaveformPaint; virtual;
    procedure WndProc(var Message: TMessage); override;
    property AutoScrollHorz: Boolean read FAutoScrollHorz write FAutoScrollHorz;
    property AutoScrollVert: Boolean read FAutoScrollVert write FAutoScrollVert;
    property Buffered: Boolean read GetBuffered write SetBuffered default True;
    property Color default clWindow;
    property Cursor default crCrossHairs;
    property DrawSelectionMode: TDrawSelectionMode read FDrawSelectionMode write SetDrawSelectionMode;
    property GridPen: TPen read FGridPen write SetGridPen;
    property Width default 149;
    property Height default 120;
    property ImageFileFormat: TImageFileFormat read FImageFileFormat write FImageFileFormat;
    property PaintMode: TSignalPaintMode read FPaintMode write SetPaintMode;
    property ParentColor default False;
    property Pen: TPen read FPen write SetPen;
    property HighlightColor: TColor read FHighlightColor write SetHighlightColor default clLime;
    property SelectionColor: TColor read FSelectionColor write SetSelectionColor default clDefault;
    property MinVertScaleWidth: Integer read FMinVertScaleWidth write SetMinVertScaleWidth;
    property MaxWindowWidth: Integer read FMaxWindowWidth write SetMaxWindowWidth; // Max(Right-Left+1)
    property MaxWindowHeight: Integer read FMaxWindowHeight write SetMaxWindowHeight; // Max(Top-Bottom+1)
    property MinWindowWidth: Integer read FMinWindowWidth write SetMinWindowWidth default 32; // Min(Right-Left+1)
    property MinWindowHeight: Integer read FMinWindowHeight write SetMinWindowHeight default 16; // Min(Top-Bottom+1)
    property MonochromeBitmap: Boolean read FMonochromeBitmap write SetMonochromeBitmap default True;
    property HorzScale: TDisplayScale read FHorzScale write SetHorzScale;
    property VertScale: TDisplayScale read FVertScale write SetVertScale;
    property HorzScrollBar: TDisplayScrollBar read FHorzScrollBar write SetHorzScrollBar;
    property VertScrollBar: TDisplayScrollBar read FVertScrollBar write SetVertScrollBar;
    property Viewport: TRect read GetViewport;
    property WindowLeft: Integer read FWindow.Left write SetWindowLeft;
    property WindowTop: Integer read FWindow.Top write SetWindowTop;
    property WindowWidth: Integer read GetWindowWidth write SetWindowWidth;
    property WindowHeight: Integer read GetWindowHeight write SetWindowHeight;
    property WorldLeft: Integer read FWorld.Left write SetWorldLeft;
    property WorldTop: Integer read FWorld.Top write SetWorldTop;
    property WorldWidth: Integer read GetWorldWidth write SetWorldWidth;
    property WorldHeight: Integer read GetWorldHeight write SetWorldHeight;
    property ZoomStartMode: TZoomStartMode read FZoomStartMode write FZoomStartMode;
    property OnGetHorzScalePoint: TGetScalePointProc read FOnGetHorzScalePoint write FOnGetHorzScalePoint;
    property OnGetVertScalePoint: TGetScalePointProc read FOnGetVertScalePoint write FOnGetVertScalePoint;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property OnSelectionChange: TNotifyEvent read FOnSelectionChange write FOnSelectionChange;
    property OnWindowChange: TNotifyEvent read FOnWindowChange write FOnWindowChange;
    property OnWorldPosChange: TPosEvent read FOnWorldPosChange write FOnWorldPosChange;
    property OnZoomMin: TNotifyEvent read FOnZoomMin write FOnZoomMin;
    property OnZoomMax: TNotifyEvent read FOnZoomMax write FOnZoomMax;
  public
    constructor Create(AOwner: TComponent); override;
    procedure BeginZoom(X, Y: Integer);
    procedure DeleteMarkersMarking(Marked: TObject);
    procedure IterateMarkers(Proc: TMarkerProc; Param: DWord);
    procedure HideHighlightedSample;
    procedure InvalidateWorldWindow;
    function DeviceX(X: TWorldCoord): Double;
    function DeviceY(Y: TWorldCoord): Double;
    procedure LineTo(X, Y: TWorldCoord);
    procedure MoveTo(X, Y: TWorldCoord);
    procedure PlotByteArray(
	const Samples: array of Byte;
	Offset: Longint;
	Channels: Integer;
	AChannel: Integer);
    procedure PlotSmallintArray(
	const Samples: array of Smallint;
	Offset: Longint;
	Channels: Integer;
	AChannel: Integer);
    procedure Plot24BitSamples(
	const Samples: array of T24BitSample;
	Offset: Longint;
	Channels: Integer;
	AChannel: Integer);
    procedure PlotIntArray(
	const Samples: array of Integer;
	Offset: Longint;
	Channels: Integer;
	AChannel: Integer);
    procedure PlotSamples(Sample: Pointer; Amplitude: TAmpFunc);
    // saves signal image.
    procedure SaveToFile(FileName: string);
    procedure SetSelection(StartValue, EndValue: Integer);
    procedure SetWindowSizeLimits(
	MinWidth, MinHeight,
	MaxWidth, MaxHeight: Integer);
    function WorldX(X: integer): TWorldCoord;
    function WorldY(Y: integer): TWorldCoord;
    property Bitmap: TBitmap read GetBitmap;
    property HighlightedSample: Integer read FHLSample write SetHighlightedSample;
    property WaveformCanvas: TCanvas read GetWaveformCanvas;
    property SelectionStart: Integer read FSelectionStart write SetSelectionStart;
    property SelectionEnd: Integer read FSelectionEnd write SetSelectionEnd;
    property Window: TRect read FWindow write SetWindow;
    property World: TRect read FWorld write SetWorld;
  published
    property Info: TComponentInfo read FInfo write SetInfo;
    property ZoomMode: TZoomMode read FZoomMode write FZoomMode;
    property OnMouseDown;
  end;

  TZoomDisplay = class(TCustomZoomDisplay)
  public
    property Viewport;
  published
    property Align;
    property AutoScrollHorz;
    property AutoScrollVert;
    property Buffered;
    property Color;
    property Cursor;
    property DrawSelectionMode;
    property Enabled;
    property GridPen;
    property HighlightColor;
    property Width;
    property Height;
    property ImageFileFormat;
    property PaintMode;
    property ParentColor;
    property Pen;
    property PopupMenu;
    property MaxWindowWidth;
    property MaxWindowHeight;
    property MinVertScaleWidth;
    property MinWindowWidth;
    property MinWindowHeight;
    property MonochromeBitmap;
    property SelectionColor;
    property HorzScale;
    property VertScale;
    property HorzScrollBar;
    property VertScrollBar;
    property WindowLeft;
    property WindowTop;
    property WindowWidth;
    property WindowHeight;
    property WorldLeft;
    property WorldTop;
    property WorldWidth;
    property WorldHeight;
    property ZoomStartMode;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnMouseLeave;
    property OnStartDrag;
    property OnMouseMove;
    property OnMouseUp;
    property OnPaint;
    property OnGetHorzScalePoint;
    property OnGetVertScalePoint;
    property OnSelectionChange;
    property OnWindowChange;
    property OnWorldPosChange;
    property OnZoomMin;
    property OnZoomMax;
  end;

  TMarkerPosEvent = procedure(Sender: TObject; var Pos: TWorldCoord) of object;
  TMarkerKind = (mkMark, mkRangeStart, mkRangeEnd);

  TMarker = class(TGraphicControl)
  private
    FMarked: TObject;
    FID: Integer;
    FKind: TMarkerKind;
    FLocked: Boolean;
    FShowPosition: Boolean;
    FVisible: Boolean;
    FPosition: TWorldCoord;
    FRepeatTimer: TTimer;
    FHighLightColor: TColor;
    FShadowColor: TColor;
    FOnPosChange: TMarkerPosEvent;
    procedure DoEndDrag(Target: TObject; X, Y: Integer); override;
    procedure DoStartDrag(var DragObject: TDragObject); override;
    procedure SetKind(Value: TMarkerKind);
    procedure SetPosition(Value: TWorldCoord);
    procedure SetShowPosition(Value: Boolean);
    procedure SetVisible(Value: Boolean);
    procedure TimerExpired(Sender: TObject);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
  protected
    destructor Destroy; override;
    procedure AssignTo(Dest: TPersistent); override;
    function HotSpot: TPoint;
    procedure MouseDown(Button: TMouseButton;
	Shift: TShiftState; X, Y: Integer); override;
    procedure ParentInvalidate; virtual; abstract;
    procedure SetInheritedVisible(Value: Boolean);
    { raises exception if AParent is no TCustomZoomDisplay }
{$IFDEF VER90}
    procedure SetParent(AParent: TWinControl); override;
{$ELSE}
    procedure ValidateContainer(AComponent: TComponent); override;
{$ENDIF}
    { Updates both world and device position }
    procedure UpdatePos; virtual; abstract;
    procedure UpdatePosition;
  public
    constructor Create(AOwner: TComponent); override;
    property Color default clBtnFace;
    property Cursor default crHand;
    property DragCursor default crGrip;
    property ID: Integer read FID write FID;
    property Kind: TMarkerKind read FKind write SetKind;
    property Locked: Boolean read FLocked write FLocked;
    property Marked: TObject read FMarked write FMarked;
    property PopupMenu;
    property Position: TWorldCoord read FPosition write SetPosition;
    property ShowHint default True;
    property ShowPosition: Boolean read FShowPosition write SetShowPosition default True;
    property Visible write SetVisible;
    property OnEndDrag;
    property OnPosChange: TMarkerPosEvent read FOnPosChange write FOnPosChange;
  end;

  THMarkerAlign = (alTop, alBottom);
  THorzMarker = class(TMarker)
  private
    FAlign: THMarkerAlign;
    procedure SetAlign(Value: THMarkerAlign);
  protected
    procedure Paint; override;
    procedure ParentInvalidate; override;
    procedure UpdatePos; override;
  public
    constructor Create(AOwner: TComponent); override;
    property Align: THMarkerAlign read FAlign write SetAlign;
  end;

  TVMarkerAlign = (alLeft, alRight);
  TVertMarker = class(TMarker)
  private
    FAlign: TVMarkerAlign;
    procedure SetAlign(Value: TVMarkerAlign);
  protected
    procedure Paint; override;
    procedure ParentInvalidate; override;
    procedure UpdatePos; override;
  public
    constructor Create(AOwner: TComponent); override;
    property Align: TVMarkerAlign read FAlign write SetAlign;
  end;

procedure Register;

implementation

{$R *.RES}
{$IFDEF Profile}
uses DsgnIntf, TypInfo, Dialogs, Z_Prof;
{$ELSE}
uses DsgnIntf, TypInfo, Dialogs;
{$ENDIF}
const
  InitRepeatPause = 400;  { pause before repeat timer (ms) }
  RepeatPause     =  50;  { repeat timer interval (ms)}
  SPlotArray = 'Plot*Array: Length of array "Samples" < WindowWidth';

function DecentContrast(Color: TColor): TColor;
var
  HLS: THLSVector;
begin
  HLS := RGBtoHLS(ColorToRGB(Color));
  if HLS.Luminance < High(THLSValue) div 3
    then Inc(HLS.Luminance, Round(0.5*(0.6875*High(THLSValue)-HLS.Luminance)))
    else Dec(HLS.Luminance, Round(0.375*HLS.Luminance));
  with HLS do Result := HLStoRGB(Hue, Luminance, Saturation);
end;

procedure VerifyWindowSize(const Value: TRect);
var
  StackTop: record end;
  Stack: record
    EBP: DWORD;
    ReturnAddress: Pointer;
  end absolute StackTop;
begin
  if (Value.Right <= Value.Left)
  or (Value.Top <= Value.Bottom) then
    raise EWindowSize.Create('Invalid window size') at Stack.ReturnAddress;
end;

constructor TMarker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Color := clBtnFace;
  Get3DColors(Color, FHighLightColor, FShadowColor);
  Cursor := crHand;
  DragCursor := crGrip;
  ShowHint := True;
  FShowPosition := True;
  FVisible := True;
  Parent := TWinControl(AOwner);
end;

destructor TMarker.Destroy;
begin
  ParentInvalidate;
  FRepeatTimer.Free;
  inherited Destroy;
end;

procedure TMarker.AssignTo(Dest: TPersistent);
var
  M: TMarker absolute Dest;
begin
  if Dest is TMarker then
  begin
    M.Color	:= Color;
    M.Cursor	:= Cursor;
    M.DragCursor := DragCursor;
    M.Hint	:= Hint;
    M.ID	:= ID;
    M.Kind	:= Kind;
    M.Locked	:= Locked;
    M.Marked	:= Marked;
    M.PopupMenu	:= PopupMenu;
    M.Position	:= Position;
    M.ShowHint	:= ShowHint;
    M.OnPosChange := OnPosChange;
  end
  else inherited AssignTo(Dest);
end;

procedure TMarker.CMColorChanged(var Message: TMessage);
begin
  Get3DColors(Color, FHighLightColor, FShadowColor);
  inherited;
end;

procedure TMarker.DoStartDrag(var DragObject: TDragObject);
var
  P: TPoint;
begin
  inherited;
  P := ClientToScreen(HotSpot);
  SetCursorPos(P.X, P.Y);
  if FRepeatTimer = nil then
  begin
    FRepeatTimer := TTimer.Create(Self);
    FRepeatTimer.OnTimer := TimerExpired;
  end;
  FRepeatTimer.Interval := InitRepeatPause;
  FRepeatTimer.Enabled  := True;
end;

procedure TMarker.DoEndDrag(Target: TObject; X, Y: Integer);
begin
  inherited;
  FRepeatTimer.Enabled  := False;
end;

function TMarker.HotSpot: TPoint;
begin
  Result.X := Width div 2;
  Result.Y := Height div 2;
end;

procedure TMarker.MouseDown(
  Button: TMouseButton;
  Shift: TShiftState;
  X, Y: Integer);
begin
  if (Button = mbLeft) and not Locked then BeginDrag(True);
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TMarker.SetKind(Value: TMarkerKind);
begin
  if Value <> FKind then
  begin
    FKind := Value;
    Invalidate;
  end;
end;
{$IFDEF VER90}
procedure TMarker.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent as TCustomZoomDisplay);
end;
{$ELSE}
procedure TMarker.ValidateContainer(AComponent: TComponent);
begin
  inherited ValidateContainer(AComponent as TCustomZoomDisplay);
end;
{$ENDIF}
procedure TMarker.SetPosition(Value: TWorldCoord);
begin
  if FPosition <> Value then
  begin
    FPosition := Value;
    if Assigned(FOnPosChange) then FOnPosChange(Self, FPosition);
    UpdatePosition;
  end;
end;

procedure TMarker.SetShowPosition(Value: Boolean);
begin
  if Value <> FShowPosition then
  begin
    FShowPosition := Value;
    ParentInvalidate;
  end;
end;

procedure TMarker.SetInheritedVisible(Value: Boolean);
begin
  inherited Visible := Value;
end;

procedure TMarker.SetVisible(Value: Boolean);
begin
  if Value <> FVisible then
  begin
    FVisible := Value;
    if Value then
    begin
      UpdatePos;
      ParentInvalidate;
    end else
    begin
      ParentInvalidate;
      SetInheritedVisible(False);
    end;
  end;
end;

procedure TMarker.TimerExpired(Sender: TObject);
begin
  FRepeatTimer.Interval := RepeatPause;
  try
    (Parent as TCustomZoomDisplay).DoAutoScroll;
  except
    FRepeatTimer.Enabled := False;
    raise;
  end;
end;

procedure TMarker.UpdatePosition;
begin
  ParentInvalidate;
  UpdatePos;
  ParentInvalidate;
end;

constructor THorzMarker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds(0, 0, 9, 15);
end;

procedure THorzMarker.Paint;
type
  TBorder = array[0..4] of TPoint;
const
  Border: array[THMarkerAlign] of TBorder = ((
	(X: 4; Y: 14),
	(X: 8; Y: 10),
	(X: 8; Y: 0),
	(X: 0; Y: 0),
	(X: 0; Y: 10)),
	(
	(X: 0; Y: 14),
	(X: 8; Y: 14),
	(X: 8; Y: 4),
	(X: 4; Y: 0),
	(X: 0; Y: 4)));

  RangeLimitSymbols: array[THMarkerAlign, Boolean, 0..2] of TPoint = (
	((
	(X: 3; Y: 8),
	(X: 3; Y: 4),
	(X: 6; Y: 4)),
	(
	(X: 5; Y: 8),
	(X: 5; Y: 4),
	(X: 2; Y: 4))),
	((
	(X: 3; Y: 6),
	(X: 3; Y: 10),
	(X: 6; Y: 10)),
	(
	(X: 5; Y: 6),
	(X: 5; Y: 10),
	(X: 2; Y: 10))));
begin
  with Canvas do
  begin
    Pen.Color := FHighLightColor;
    Brush.Color := Color;
    Polygon(Border[FAlign]);
    Pen.Color := FShadowColor;
    PolyLine(Slice(Border[FAlign], 3));
    if FKind = mkMark then Exit;
    Pen.Mode := pmNot;
    PolyLine(RangeLimitSymbols[FAlign, FKind = mkRangeEnd]);
    Pen.Mode := pmCopy;
  end;
end;

procedure THorzMarker.ParentInvalidate;
var
  R: TRect;
  Z: TCustomZoomDisplay;
begin
  if Visible and ShowPosition and (Parent <> nil) then
  begin
    Z := TCustomZoomDisplay(Parent);
    R := Z.Viewport;
    R.Left   := Left + HotSpot.X;
    R.Right  := R.Left+1;
    if FAlign = alTop
      then Inc(R.Top, Height)
      else Dec(R.Bottom, Height);
    with Z do
    if Buffered then
    with WaveformCanvas do
    begin
      Pen.Mode := pmNot;
      MoveTo(R.Left-FViewport.X, 0);
      LineTo(R.Left-FViewport.X, FViewport.H+1);
    end;
    InvalidateRect(Parent.Handle, @R, True);
  end;
end;

procedure THorzMarker.SetAlign(Value: THMarkerAlign);
begin
  if Value <> FAlign then
  begin
    FAlign := Value;
    if Value = alTop
      then SetBounds(Left, 0, Width, Height)
      else SetBounds(Left, TCustomZoomDisplay(Parent).FViewport.H+1-Height, Width, Height);
  end;
end;

procedure THorzMarker.UpdatePos;
var
  Y: Integer;
  Z: TCustomZoomDisplay;
begin
  Z := TCustomZoomDisplay(Parent);
  if (Position > Z.Window.Right)
  or (Position < Z.Window.Left) then
  begin
    SetInheritedVisible(False);
  end else
  begin
    if FAlign = alTop
      then Y := Z.FViewport.Y
      else Y := Z.FViewport.Y + Z.FViewport.H+1-Height;
    SetBounds(Z.FViewport.X+Round(Z.DeviceX(Position))-HotSpot.X, Y, Width, Height);
    if FVisible then SetInheritedVisible(True);
  end;
end;

constructor TVertMarker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds(0, 0, 15, 9);
end;

procedure TVertMarker.Paint;
type
  TBorder = array[0..4] of TPoint;
const
  Border: array[TVMarkerAlign] of TBorder = ((
	(X: 14; Y: 4),
	(X: 10; Y: 8),
	(X:  0; Y: 8),
	(X:  0; Y: 0),
	(X: 10; Y: 0)),
	(
	(X: 14; Y: 0),
	(X: 14; Y: 8),
	(X:  4; Y: 8),
	(X:  0; Y: 4),
	(X:  4; Y: 0)));

  RangeLimitSymbols: array[TVMarkerAlign, Boolean, 0..2] of TPoint = (
	((
	(X: 8; Y: 5),
	(X: 4; Y: 5),
	(X: 4; Y: 2)),
	(
	(X: 8; Y: 3),
	(X: 4; Y: 3),
	(X: 4; Y: 6))),
	((
	(X:  6; Y: 5),
	(X: 10; Y: 5),
	(X: 10; Y: 2)),
	(
	(X:  6; Y: 3),
	(X: 10; Y: 3),
	(X: 10; Y: 6))));
begin
  with Canvas do
  begin
    Pen.Color := FHighLightColor;
    Brush.Color := Color;
    Polygon(Border[FAlign]);
    Pen.Color := FShadowColor;
    PolyLine(Slice(Border[FAlign], 3));
    if FKind = mkMark then Exit;
    Pen.Mode := pmNot;
    PolyLine(RangeLimitSymbols[FAlign, FKind = mkRangeEnd]);
    Pen.Mode := pmCopy;
  end;
end;

procedure TVertMarker.ParentInvalidate;
var
  R: TRect;
begin
  if Visible and ShowPosition and (Parent <> nil) then
  begin
    R.Top    := Top + HotSpot.Y;
    R.Bottom := R.Top+1;
    if FAlign = alLeft then
    begin
      R.Left   := Width;
      R.Right  := Parent.Width;
    end else
    begin
      R.Left   := 0;
      R.Right  := Parent.Width-Width;
    end;
    with TCustomZoomDisplay(Parent) do
    if Buffered then
    with WaveformCanvas do
    begin
      Pen.Mode := pmNot;
      MoveTo(R.Top, 0);
      LineTo(R.Top, FViewport.W+1);
    end;
    InvalidateRect(Parent.Handle, @R, True);
  end;
end;

procedure TVertMarker.SetAlign(Value: TVMarkerAlign);
begin
  if Value <> FAlign then
  begin
    FAlign := Value;
    if Value = alLeft
      then SetBounds(0, Top, Width, Height)
      else SetBounds(TCustomZoomDisplay(Parent).FViewport.W+1-Width, Top, Width, Height);
  end;
end;

procedure TVertMarker.UpdatePos;
var
  X: Integer;
  Z: TCustomZoomDisplay;
begin
  Z := TCustomZoomDisplay(Parent);
  if (Position > Z.Window.Top)
  or (Position < Z.Window.Bottom) then
  begin
    SetInheritedVisible(False);
  end else
  begin
    if FAlign = alLeft
      then X := Z.FViewport.X
      else X := Z.FViewport.X+Z.FViewport.W+1-Width;
    SetBounds(X, Z.FViewport.Y+Round(Z.DeviceY(Position))-HotSpot.Y, Width, Height);
    if FVisible then SetInheritedVisible(True);
  end;
end;

constructor TDisplayScale.Create(ADisplay: TCustomZoomDisplay);
begin
  inherited Create;
  FDisplay := ADisplay;
end;

procedure TDisplayScale.Changed;
begin
  if Assigned(FDisplay) then FDisplay.InvalidateWorldWindow;
  inherited Changed;
end;

procedure TDisplayScale.SetShowGrid(Value: Boolean);
begin
  if Value <> FShowGrid then
  begin
    FShowGrid := Value;
    Changed;
  end;
end;

constructor TDisplayScrollBar.Create(ADisplay: TCustomZoomDisplay;
  AKind: TScrollBarKind);
begin
  inherited Create;
  FDisplay := ADisplay;
  FKind := AKind;
  FPageFraction := 10;
  FVisible := True;
end;

procedure TDisplayScrollBar.Assign(Source: TPersistent);
begin
  if Source is TDisplayScrollBar then
  begin
    Visible := TDisplayScrollBar(Source).Visible;
  end
  else inherited Assign(Source);
end;

function TDisplayScrollBar.GetMask(Flags: UInt): UInt;
const
  ScrollInfoMask: array[Boolean] of UInt = (0, SIF_DISABLENOSCROLL);
begin
  Result := ScrollInfoMask[FPersistent] or Flags;
end;

function TDisplayScrollBar.GetMin: Integer;
begin
  if FKind = sbHorizontal
    then Result := FDisplay.FWorld.Left
    else Result := -FDisplay.FWorld.Top;
end;

function TDisplayScrollBar.GetMax: Integer;
begin
  with FDisplay do
  if FKind = sbHorizontal
    then Result := FWorld.Right
    else Result := -FWorld.Bottom;
end;

function TDisplayScrollBar.GetPageSize: Integer;
begin
  if FVisible then
    if FKind = sbHorizontal
      then Result := FDisplay.WindowWidth
      else Result := FDisplay.WindowHeight
  else Result := Max-Min+1;
end;

function TDisplayScrollBar.GetPosition: Integer;
begin
  with FDisplay do
  if FKind = sbHorizontal
    then Result := FXPos
    else Result := FYPos;
end;

function TDisplayScrollBar.GetScrollPos: Integer;
begin
  with FDisplay do
  if FKind = sbHorizontal
    then Result := FXPos
    else Result := 1-FYPos-PageSize;
end;

procedure TDisplayScrollBar.SetPersistent(Value: Boolean);
var
  Code: Integer;
  ScrollInfo: TScrollInfo;
begin
  if Value <> FPersistent then
  begin
    ScrollInfo.cbSize := SizeOf(ScrollInfo);
    ScrollInfo.fMask := SIF_ALL;
    Code := SB_HORZ;
    if FKind = sbVertical then Code := SB_VERT;
    GetScrollInfo(FDisplay.Handle, Code, ScrollInfo);
    FPersistent := Value;
    ScrollInfo.fMask := GetMask(ScrollInfo.fMask);
    SetScrollInfo(FDisplay.Handle, Code, ScrollInfo, True);
  end;
end;

procedure TDisplayScrollBar.SetPosition(Value: Integer);
begin
  with FDisplay do
  if FKind = sbHorizontal
    then ScrollTo(Value, FYPos)
    else ScrollTo(FXPos, Value);
end;

procedure TDisplayScrollBar.SetVisible(Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    FDisplay.UpdateScrollBars;
  end;
end;

procedure TDisplayScrollBar.Update;
var
  Code: Word;
  ScrollInfo: TScrollInfo;
begin
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.fMask := GetMask(SIF_ALL);
  if Kind = sbHorizontal
    then Code := SB_HORZ
    else Code := SB_VERT;
  ScrollInfo.nMin := Min;
  ScrollInfo.nMax := Max;
  ScrollInfo.nPage := PageSize;
  ScrollInfo.nPos := GetScrollPos;
  ScrollInfo.nTrackPos := GetScrollPos;
  SetScrollInfo(FDisplay.Handle, Code, ScrollInfo, True);
end;

{ TCustomZoomDisplay }

constructor TCustomZoomDisplay.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  FPen := TPen.Create;
  FPen.OnChange := PenChanged;
  FGridPen := TPen.Create;
  FGridPen.OnChange := GridPenChanged;
  FViewport.dW := 1;
  FViewport.dH := 1;
  SetBounds(0, 0, 149, 120);
  FHorzScrollBar := TDisplayScrollBar.Create(Self, sbHorizontal);
  FVertScrollBar := TDisplayScrollBar.Create(Self, sbVertical);
  FWorld.Right := 100;
  FWorld.Top := 100;
  FWindow := FWorld;
  FYPos := FWindow.Bottom;
  SetWindowSizeLimits(32, 16, FWorld.Right+1, FWorld.Top-FWorld.Bottom+1);
  FSelectionColor := clDefault;
  Color := clWindow;
  Cursor := crCrossHairs;
  ParentColor := False;
  FDoDrawMarkerPositions := True;
  FHighlightColor := clLime;
  FHorzScale := TDisplayScale.Create(Self);
  FVertScale := TDisplayScale.Create(Self);
  FHorzScale.OnDrawGrid := ScaleDrawGrid;
  FVertScale.OnDrawGrid := ScaleDrawGrid;
  FVertScale.BeginUpdate;
  try
    FVertScale.TextPosition := tpTopLeft;
    FVertScale.StartPointX := -1;
    FVertScale.EndPointX := -1;
  finally
    FVertScale.EndUpdate;
  end;
  FMonochromeBitmap := True;
  Buffered := True;
  FInfo := TComponentInfo.Create(Self,
           CopyrightOwner,
	   '1993-1999',
           VersionIntToStr(Version),
           ComponentLicense,
           WebsiteURL,
           MailAddr);
end;

destructor TCustomZoomDisplay.Destroy;
begin
  FInfo.Free;
  FPen.Free;
  FGridPen.Free;
  FBitmap.Free;
  FHorzScrollBar.Free;
  FVertScrollBar.Free;
  inherited Destroy;
end;

procedure TCustomZoomDisplay.CheckFileImageVersion;
const
  DefaultChanged =
    'File image version of %s.%s is %s.'#13#13+
    'Note that the default values for the MaxWindowWidth and '+
    'MaxWindowHeight properties have been changed in version 3.2.'#13#13+
    '%0:s tries to compensate for this, but will fail in doing so if '+
    'the value stored on disk accidentally equals the new default value.'#13#13+
    'Hence it is recommended to check the values of MaxWindowWidth, '+
    'MaxWindowHeight, WindowWidth and WindowHeight on correctness.';
var
  SVer: string;
  FileImageVersion: Integer;
begin
  try
    FileImageVersion := VersionStrToInt(FInfo.FileImageVersion);
  except
    on EConvertError do FileImageVersion := 0;
  end;
  SVer := FInfo.FileImageVersion;
  if SVer = '' then SVer := '< 3.2';
  if FileImageVersion < $03020000 then // < 3.2
  begin
    if (MaxWindowWidth = 101)
    or (MaxWindowHeight = 101) then
    begin
      MessageDlg(Format(DefaultChanged, [Owner.Name, Name, SVer]), mtWarning, [mbOK], 0);
      if (MaxWindowWidth  = 101) then // 101: new default value
      begin
        MaxWindowWidth  := WorldWidth;
        WindowWidth := WorldWidth;
      end;
      if (MaxWindowHeight = 101) then // 101: new default value
      begin
        MaxWindowHeight := WorldHeight;
        WindowHeight := WorldHeight;
      end;
    end;
  end;
end;

procedure TCustomZoomDisplay.DoMouseLeave;
begin
  if FMouseOver then
  begin
    FMouseOver := False;
    if Assigned(FOnMouseLeave) then FOnMouseLeave(Self);
  end;
end;

procedure TCustomZoomDisplay.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  // might be sent by a child control when it is left
  if (Message.LParam = 0) then DoMouseLeave;
end;

procedure TCustomZoomDisplay.WMNCMouseMove(var Message: TMessage);
// mouse left client area
begin
  inherited;
  DoMouseLeave;
end;

procedure TCustomZoomDisplay.CMColorChanged(var Message: TMessage);
begin
  inherited;
  FDecentContrast := DecentContrast(Color);
  UpdateSelectionBrush;
  if FBitmap <> nil then
{$IFNDEF VER90}
    if FBitmap.Monochrome and (FBitmap.Handle <> 0) then
      SetBitmapColors(FBitmap, [Color, FPen.Color], 0)
    else
{$ENDIF}
    InvalidateWorldWindow;
end;

procedure TCustomZoomDisplay.GridPenChanged(Sender: TObject);
begin
  InvalidateWorldWindow;
end;

procedure TCustomZoomDisplay.PenChanged(Sender: TObject);
begin
{$IFNDEF VER90}
  if (FBitmap <> nil) and FBitmap.Monochrome then
    SetBitmapColors(FBitmap, [Color, FPen.Color], 0);
{$ENDIF}
  WaveformCanvas.Pen := FPen;
  InvalidateWorldWindow;
end;

procedure TCustomZoomDisplay.DoAutoScroll;
var
  CursorPos: TPoint;
  R: TRect;
  Dx, Dy: Integer;
begin
  if FAutoScrollVert
  or FAutoScrollHorz then
  begin
    GetCursorPos(CursorPos);
    CursorPos := ScreenToClient(CursorPos);
    R := Viewport;
    Dx := 0; Dy := 0;
    if FAutoScrollVert then
      if CursorPos.Y < R.Top then
	Dy := iMin(-FYIncrement, iMax(-WindowHeight, ((CursorPos.Y-R.Top) div 10) * FYIncrement)) else
      if CursorPos.Y > R.Bottom then
	Dy := iMax(FYIncrement, iMin(WindowHeight, ((CursorPos.Y-R.Bottom) div 10) * FYIncrement));
    if FAutoScrollHorz then
      if CursorPos.X < R.Left then
	Dx := iMin(-FXIncrement, iMax(-WindowWidth, ((CursorPos.X-R.Left) div 10) * FXIncrement)) else
      if CursorPos.X > R.Right then
	Dx := iMax(FXIncrement, iMin(WindowWidth, ((CursorPos.X-R.Right) div 10) * FXIncrement));
    ScrollBy(Dx, Dy);
  end;
end;

procedure TCustomZoomDisplay.CreateWnd;
begin
  inherited CreateWnd;
  UpdateScrollBars;
end;

procedure TCustomZoomDisplay.Loaded;
begin
  inherited;
  CheckFileImageVersion;
  FXPos := FWindow.Left;
  FYPos := FWindow.Bottom;
  with FViewport do SetViewport(X, Y, ClientWidth-dW+1, ClientHeight-dH+1);
end;

procedure TCustomZoomDisplay.Error(Text: string);
var
  StackTop: record end;
  Stack: record
    EBP: Integer;
    ReturnAddress: Pointer;
  end absolute StackTop;
  C: string;
begin
  C := ClassName;
  C := Copy(C, 2, Length(C)-1);
  raise EZoomDisplayError.CreateFmt('%s %s: %s',
    [C, Name, Text]) at Stack.ReturnAddress;
end;

procedure TCustomZoomDisplay.InvalidateWorldWindow;
begin
  if not FInWaveformPaint then
  if not FWindowInvalid then
  begin
    FWindowInvalid := True;
    Invalidate;
  end;
end;

function TCustomZoomDisplay.GetBitmap: TBitmap;
begin
  if FBitmap = nil then
  begin
    Buffered := True;
    WaveformPaint;
  end;
  Result := FBitmap;
end;

function TCustomZoomDisplay.GetBuffered: Boolean;
begin
  Result := FBitmap <> nil;
end;

procedure TCustomZoomDisplay.SetBuffered(Value: Boolean);
begin
  if Value <> Buffered then
  begin
    if Value then
    begin
      FBitmap := TBitmap.Create;
      if HandleAllocated then
      begin
        FBitmap.Width  := ClientWidth;
        FBitmap.Height := ClientHeight;
      end;
      FBitmap.Monochrome := FMonochromeBitmap;
      FRestoreColors := True;
      FBitmapInvalid := True;
    end else
    begin
      FBitmap.Free;
      FBitmap := nil;
    end;
    InvalidateWorldWindow;
  end;
end;

function TCustomZoomDisplay.GetPalette: HPalette;
begin
  if FBitmap <> nil
    then Result := FBitmap.Palette
    else Result := 0;
end;

procedure TCustomZoomDisplay.SetHorzScale(Value: TDisplayScale);
begin
  FHorzScale.Assign(Value);
end;

procedure TCustomZoomDisplay.SetVertScale(Value: TDisplayScale);
begin
  FVertScale.Assign(Value);
end;

procedure TCustomZoomDisplay.SetPen(Value: TPen);
begin
  FPen.Assign(Value);
end;

procedure TCustomZoomDisplay.SetGridPen(Value: TPen);
begin
  FPen.Assign(Value);
end;

procedure TCustomZoomDisplay.SetHorzScrollBar(Value: TDisplayScrollBar);
begin
  FHorzScrollBar.Assign(Value);
end;

procedure TCustomZoomDisplay.SetVertScrollBar(Value: TDisplayScrollBar);
begin
  FVertScrollBar.Assign(Value);
end;

procedure TCustomZoomDisplay.UpdateScrollBars;
begin
  if HandleAllocated then
  begin
    FVertScrollBar.Update;
    FHorzScrollBar.Update;
  end;
end;

procedure TCustomZoomDisplay.SetSelection(StartValue, EndValue: Integer);
var
  Temp: Integer;
  R: TRect;
begin
  if StartValue < FWorld.Left then StartValue := FWorld.Left else
  if StartValue > FWorld.Right then StartValue := FWorld.Right;
  if EndValue < FWorld.Left then EndValue := FWorld.Left else
  if EndValue > FWorld.Right then EndValue := FWorld.Right;
  if StartValue > EndValue then // swap
  begin
    Temp := StartValue;
    StartValue := EndValue;
    EndValue := Temp;
  end;
  if (StartValue <> FSelectionStart)
  or (EndValue <> FSelectionEnd) then
  begin
    R := GetSelectionRect;
    if (FBitmap <> nil) and not FBitmapInvalid then DrawSelection;
    with FViewport do OffsetRect(R, X, Y);
    InvalidateRect(Handle, @R, False);
    FSelectionStart := StartValue;
    FSelectionEnd := EndValue;
    R := GetSelectionRect;
    if (FBitmap <> nil) and not FBitmapInvalid then DrawSelection;
    with FViewport do OffsetRect(R, X, Y);
    InvalidateRect(Handle, @R, False);
    if Assigned(FOnSelectionChange) then FOnSelectionChange(Self);
  end;
end;

procedure TCustomZoomDisplay.SetSelectionStart(Value: Integer);
begin
  SetSelection(Value, SelectionEnd);
end;

procedure TCustomZoomDisplay.SetSelectionEnd(Value: Integer);
begin
  SetSelection(SelectionStart, Value);
end;

function TCustomZoomDisplay.GetViewport: TRect;
begin
  with FViewport do
    Result := Rect(X, Y, X+W+1, Y+H+1);
end;

procedure TCustomZoomDisplay.SetViewport(Left, Top, Width, Height: Integer);
begin
  with FViewport do
  if (Left <> X) or (Width  <> W+1)
  or (Top  <> Y) or (Height <> H+1) then
  begin
    X := Left;
    Y := Top;
    if Left+Width > ClientWidth then Width := ClientWidth-Left;
    if Top+Height > ClientHeight then Height := ClientHeight-Top;
    dW := ClientWidth-Width+1;
    dH := ClientHeight-Height+1;
    FViewport.W := ClientWidth-FViewport.dW;
    FViewport.H := ClientHeight-FViewport.dH;
    FVertScale.StartPointY := Height-1;
    FHorzScale.BeginUpdate;
    try
      FHorzScale.StartPointY := Height;
      FHorzScale.EndPoint := Point(Width, Height);
    finally
      FHorzScale.EndUpdate;
    end;
  end;
end;

procedure TCustomZoomDisplay.SetWindow(Value: TRect);
begin
  VerifyWindowSize(Value);
  BoundWindow(Value);
  if not EqualRect(FWindow, Value) then
  begin
    FWindow := Value;
    FXPos := FWindow.Left;
    FYPos := FWindow.Bottom;
    ScrollTo(FWindow.Left, FWindow.Bottom);
    InvalidateWorldWindow;
  end;
end;

procedure TCustomZoomDisplay.SetWorld(const Value: TRect);
var
  EqualsWH: Boolean;
  EqualsWW: Boolean;
  EqualsMaxWH: Boolean;
  EqualsMaxWW: Boolean;
  W, H: Integer;
begin
  VerifyWindowSize(Value);
  if not EqualRect(Value, FWorld) then
  begin
    EqualsWH := WorldHeight = WindowHeight;
    EqualsWW := WorldWidth = WindowWidth;
    EqualsMaxWH := WorldHeight = MaxWindowHeight;
    EqualsMaxWW := WorldWidth = MaxWindowWidth;
    FWorld := Value;
    if not (csReading in ComponentState) then
    begin
      W := Value.Right-Value.Left+1;
      H := Value.Top-Value.Bottom+1;
      if EqualsMaxWW or (MaxWindowWidth > W) then MaxWindowWidth := W;
      if EqualsMaxWH or (MaxWindowHeight > H) then MaxWindowHeight := H;
      if EqualsWH then WindowHeight := H;
      if EqualsWW then WindowWidth := W;
      BoundWindow(FWindow);
    end;
  end;
end;

function TCustomZoomDisplay.GetWindowWidth: Integer;
begin
  Result := FWindow.Right-FWindow.Left+1;
end;

function TCustomZoomDisplay.GetWindowHeight: Integer;
begin
  Result := FWindow.Top-FWindow.Bottom+1;
end;

function TCustomZoomDisplay.GetWorldWidth: Integer;
begin
  Result := FWorld.Right-FWorld.Left+1;
end;

function TCustomZoomDisplay.GetWorldHeight: Integer;
begin
  Result := FWorld.Top-FWorld.Bottom+1;
end;

procedure TCustomZoomDisplay.SetWindowSizeLimits(
  MinWidth, MinHeight,
  MaxWidth, MaxHeight: Integer);
var
  MaxW, MaxH: Integer;
begin
  if not (csReading in ComponentState) then
  begin
    MaxW := FWorld.Right-FWorld.Left+1;
    MaxH := FWorld.Top-FWorld.Bottom+1;
    if MaxWidth > MaxW then MaxWidth := MaxW;
    if MaxHeight > MaxH then MaxHeight := MaxH;
    if MinWidth > MaxWidth then MinWidth := MaxWidth;
    if MinHeight > MaxHeight then MinHeight := MaxHeight;
  end;
  FMinWindowWidth := MinWidth;
  FMinWindowHeight := MinHeight;
  FMaxWindowWidth := MaxWidth;
  FMaxWindowHeight := MaxHeight;
  if not (csReading in ComponentState) then
  if BoundWindow(FWindow) then InvalidateWorldWindow;
end;

procedure TCustomZoomDisplay.SetMinWindowWidth(Value: Integer);
begin
  SetWindowSizeLimits(Value, MinWindowHeight, MaxWindowWidth, MaxWindowHeight);
end;

procedure TCustomZoomDisplay.SetMinWindowHeight(Value: Integer);
begin
  SetWindowSizeLimits(MinWindowWidth, Value, MaxWindowWidth, MaxWindowHeight);
end;

procedure TCustomZoomDisplay.SetMaxWindowWidth(Value: Integer);
begin
  SetWindowSizeLimits(MinWindowWidth, MinWindowHeight, Value, MaxWindowHeight);
end;

procedure TCustomZoomDisplay.SetMaxWindowHeight(Value: Integer);
begin
  SetWindowSizeLimits(MinWindowWidth, MinWindowHeight, MaxWindowWidth, Value);
end;

procedure TCustomZoomDisplay.SetWindowLeft(Value: Integer);
var
  W: TRect;
begin
  W := FWindow;
  W.Right := W.Right+Value-W.Left;
  W.Left := Value;
  SetWindow(W);
end;

procedure TCustomZoomDisplay.SetWindowWidth(Value: Integer);
var
  W: TRect;
begin
  W := FWindow;
  W.Right := W.Left+Value-1;
  SetWindow(W);
end;

procedure TCustomZoomDisplay.SetWindowTop(Value: Integer);
var
  W: TRect;
begin
  W := FWindow;
  W.Bottom := W.Bottom+Value-W.Top;
  W.Top := Value;
  SetWindow(W);
end;

procedure TCustomZoomDisplay.SetWindowHeight(Value: Integer);
var
  W: TRect;
begin
  W := FWindow;
  W.Bottom := W.Top-Value+1;
  SetWindow(W);
end;

procedure TCustomZoomDisplay.SetWorldLeft(Value: Integer);
var
  W: TRect;
begin
  W := FWorld;
  W.Right := W.Right+Value-W.Left;
  W.Left := Value;
  SetWorld(W);
end;

procedure TCustomZoomDisplay.SetWorldWidth(Value: Integer);
var
  W: TRect;
begin
  W := FWorld;
  W.Right := W.Left+Value-1;
  SetWorld(W);
end;

procedure TCustomZoomDisplay.SetWorldTop(Value: Integer);
var
  W: TRect;
begin
  W := FWorld;
  W.Bottom := W.Bottom+Value-W.Top;
  W.Top := Value;
  SetWorld(W);
end;

procedure TCustomZoomDisplay.SetWorldHeight(Value: Integer);
var
  W: TRect;
begin
  W := FWorld;
  W.Bottom := W.Top-Value+1;
  SetWorld(W);
end;

function TCustomZoomDisplay.BoundWindow(var AWindow: TRect): Boolean;
var
  Dx, Dy: Integer;
  OldWindow: TRect;
begin
  Result := False;
  if not (csLoading in ComponentState) then
  with AWindow do
  begin
    OldWindow := AWindow;
    if Right-Left < MinWindowWidth-1 then Right := Left + MinWindowWidth-1 else
    if Right-Left > MaxWindowWidth-1 then Right := Left + MaxWindowWidth-1;
    if Top-Bottom < MinWindowHeight-1 then Top := Bottom + MinWindowHeight-1 else
    if Top-Bottom > MaxWindowHeight-1 then Top := Bottom + MaxWindowHeight-1;
    Dx := 0;
    Dy := 0;
    if Right > FWorld.Right then Dx := FWorld.Right-Right else
    if Left < FWorld.Left then Dx := FWorld.Left-Left;
    if Top > FWorld.Top then Dy := FWorld.Top-Top else
    if Bottom < FWorld.Bottom then Dy := FWorld.Bottom-Bottom;
    OffsetRect(AWindow, Dx, Dy);
    Result := not EqualRect(AWindow, OldWindow);
    if Result and (@AWindow = @FWindow) then
    begin
      FXPos := AWindow.Left;
      FYPos := AWindow.Bottom;
    end;
  end;
end;

function TCustomZoomDisplay.DevicePositionCode(P: TDoublePoint): TClipCode;
begin
  Result := [ccCenter];
  if P.Y < 0 then Include(Result, ccTop) else
  if P.Y > FViewport.H then Include(Result, ccBottom);
  if P.X > FViewport.W then Include(Result, ccRight) else
  if P.X < 0 then Include(Result, ccLeft);
end;

function TCustomZoomDisplay.DeviceX(X: TWorldCoord): Double;
begin
  Result := (X-FWindow.Left)*FViewport.W/(FWindow.Right-FWindow.Left);
end;

function TCustomZoomDisplay.DeviceY(Y: TWorldCoord): Double;
begin
  Result := (Y-FWindow.Top)*FViewport.H/(FWindow.Bottom-FWindow.Top);
end;

procedure TCustomZoomDisplay.DoLineTo(DC: hDC; X, Y: TWorldCoord);
var
  PrevCode, Code1, Code2: TClipCode;
  P, Q: TDoublePoint;
  OldPos: TPoint;
  Qx, Qy: Integer;

  procedure Clip(Code: TClipCode; var P: TDoublePoint; Q: TDoublePoint); near;
  begin
    if ccTop in Code then
    begin
      P.X := P.X - P.Y*(Q.X-P.X)/(Q.Y-P.Y);
      P.Y := 0;
    end else
    if ccBottom in Code then
    begin
      P.X := P.X + (FViewport.H-P.Y)*(Q.X-P.X)/(Q.Y-P.Y);
      P.Y := FViewport.H;
    end else
    if ccRight in Code then
    begin
      P.Y := P.Y + (FViewport.W-P.X)*(Q.Y-P.Y)/(Q.X-P.X);
      P.X := FViewport.W;
    end else
    if ccLeft in Code then
    begin
      P.Y := P.Y - P.X*(Q.Y-P.Y)/(Q.X-P.X);
      P.X := 0;
    end;
  end;

begin
  P := FCurrentPos;
  Q.X := DeviceX(X);
  Q.Y := DeviceY(Y);
  FCurrentPos := Q;
  PrevCode := FCurrentCode;
  Code1 := PrevCode;
  Code2 := DevicePositionCode(Q);
  FCurrentCode := Code2;
  repeat
    if Code1 * Code2 <> [ccCenter] then exit
    else if Code1 + Code2 = [ccCenter] then
    begin
      if PrevCode <> [ccCenter] then MoveToEx(DC, Round(P.X), Round(P.Y), @OldPos);
      Qx := Round(Q.X);
      Qy := Round(Q.Y);
      Windows.LineTo(DC, Qx, Qy);
      // Draw the end point if neccessary
      if FCurrentCode <> [ccCenter] then Windows.LineTo(DC, Qx+1, Qy);
      exit;
    end;
    if Code1 = [ccCenter] then Clip(Code2, Q, P) else Clip(Code1, P, Q);
    Code1 := DevicePositionCode(P);
    Code2 := DevicePositionCode(Q);
  until False;
end;

procedure TCustomZoomDisplay.LineTo(X, Y: TWorldCoord);
begin
  DoLineTo(WaveformCanvas.Handle, X, Y);
end;

procedure TCustomZoomDisplay.MoveTo(X, Y: TWorldCoord);
begin
  FCurrentPos.X := DeviceX(X);
  FCurrentPos.Y := DeviceY(Y);
  FCurrentCode := DevicePositionCode(FCurrentPos);
  if FCurrentCode = [ccCenter] then
    WaveformCanvas.MoveTo(Round(FCurrentPos.X), Round(FCurrentPos.Y));
end;

procedure TCustomZoomDisplay.MoveWindow(X, Y: Integer);
begin
  OffsetRect(FWindow, X-FWindow.Left, Y-FWindow.Bottom);
  BoundWindow(FWindow);
end;
{
function TCustomZoomDisplay.PositionCode(P: TPoint): TClipCode;
begin
  Result := [ccCenter];
  if P.Y > FWindow.Top    then Include(Result, ccTop) else
  if P.Y < FWindow.Bottom then Include(Result, ccBottom);
  if P.X > FWindow.Right  then Include(Result, ccRight) else
  if P.X < FWindow.Left   then Include(Result, ccLeft);
end;
}
procedure TCustomZoomDisplay.CMDrag(var Message: TCMDrag);
begin
  if Message.DragMessage = dmFindTarget
    then Message.Result := Longint(Self)
    else inherited;
end;

procedure TCustomZoomDisplay.DragDrop(
  Source: TObject;
  X, Y: Integer);
begin
  if Source is THorzMarker
    then TMarker(Source).Position := WorldX(X-FViewport.X) else
  if Source is TVertMarker
    then TMarker(Source).Position := WorldY(Y-FViewport.Y)
  else inherited;
end;

procedure TCustomZoomDisplay.HorzDragLine(
  X, Y: Integer;
  DragLineVisible: Boolean);
begin
  Canvas.Pen.Mode := pmNot;
  if FDragLineVisible then
  begin
    Canvas.MoveTo(FDragPos.X, 0);
    Canvas.LineTo(FDragPos.X, FViewport.H+1);
  end;
  FDragLineVisible := DragLineVisible;
  if not DragLineVisible then Exit;
  FDragPos.X := X;
  FDragPos.Y := Y;
  Canvas.MoveTo(X, 0);
  Canvas.LineTo(X, FViewport.H+1);
end;

procedure TCustomZoomDisplay.VertDragLine(
  X, Y: Integer;
  DragLineVisible: Boolean);
begin
  Canvas.Pen.Mode := pmNot;
  if FDragLineVisible then
  begin
    Canvas.MoveTo(0,     FDragPos.Y);
    Canvas.LineTo(Width, FDragPos.Y);
  end;
  FDragLineVisible := DragLineVisible;
  if not DragLineVisible then Exit;
  FDragPos.X := X;
  FDragPos.Y := Y;
  Canvas.MoveTo(0,     Y);
  Canvas.LineTo(Width, Y);
end;

procedure TCustomZoomDisplay.DragOver(
  Source: TObject;
  X, Y: Integer;
  State: TDragState;
  var Accept: Boolean);
var
  DragLineVisible: Boolean;

  procedure DragLine;
  begin
    if Source is THorzMarker
      then HorzDragLine(X, Y, DragLineVisible)
      else VertDragLine(X, Y, DragLineVisible);
  end;

begin
  DragLineVisible := False;
  if PtInRect(Viewport, Point(X, Y)) then
  begin
    inherited;
    WorldPosChanged(X, Y);
    if State = dsDragLeave then DoMouseLeave else
    begin
      FMouseOver := True;
      if Accept then
      begin
        DragLineVisible := True;
{ The following code has the effect of avoiding some dirt
  otherwise left behind by *DragLine when dragging is
  started by clicking in a TCustomZoomDisplay placed in a
  window which is partially covered by another window. }
        if Source is TMarker then with TMarker(Source) do
        if Visible and (Parent = Self) then
        if Source is THorzMarker then
        begin
          if Left+HotSpot.X = X then DragLineVisible := False;
        end else
          if Top +HotSpot.Y = Y then DragLineVisible := False;
      end;
    end;
    DragLine;
  end else
  begin
    DragLine;
    DoMouseLeave;
    Accept := False;
  end;
end;

procedure TCustomZoomDisplay.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if PtInRect(Viewport, Point(X, Y))
    then FMouseOver := True
    else DoMouseLeave;
  WorldPosChanged(X, Y);
  if FZooming then ContinueZoom(X, Y);
end;

procedure TCustomZoomDisplay.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then
  if not (csDesigning in ComponentState) then
  if FZoomMode <> zmZoomOff then
  if FZoomStartMode = zmAutomatic then
  begin
    BeginZoom(X, Y);
    Exit;
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TCustomZoomDisplay.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) and FZooming then
  begin
    FAlternateZoomModes := ssShift in Shift;
    EndZoom;
  end;
  inherited;
end;

procedure TCustomZoomDisplay.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    wm_EraseBkgnd:
	begin
	  Message.Result := 1;
	  Exit;
	end;
    wm_Paint: UpdateWorldWindow;
  end;
  inherited WndProc(Message);
end;

procedure TCustomZoomDisplay.DrawMarkerPositions;
var
  i, X, Y: Integer;
  M: TMarker;
begin
  if FDoDrawMarkerPositions then
  with WaveformCanvas do
  begin
    Pen.Mode := pmNot;
    for i := 0 to ControlCount-1 do
    begin
      M := TMarker(Controls[i]);
      if M.Visible and M.ShowPosition then
      if M is THorzMarker then
      begin
        X := Round(DeviceX(M.Position));
        MoveTo(X, 0);
        LineTo(X, FViewport.H+1);
      end else
      if M is TVertMarker then
      begin
        Y := Round(DeviceY(M.Position));
        MoveTo(0, Y);
        LineTo(FViewport.W+1, Y);
      end;
    end;
  end;
end;

procedure TCustomZoomDisplay.SetSelectionColor(Value: TColor);
begin
  if (FDrawSelectionMode = smPatInvert) and
  ((Value <> FSelectionColor) or (csLoading in ComponentState)) then
  begin
    DrawSelection;
    FSelectionColor := Value;
    UpdateSelectionBrush;
    DrawSelection;
    Invalidate;
  end
  else FSelectionColor := Value;
end;

procedure TCustomZoomDisplay.UpdateSelectionBrush;
begin
  if FSelBrush <> nil then
  if FSelectionColor = clDefault
    then FSelBrush.Color := FDecentContrast xor ColorToRGB(Color)
    else FSelBrush.Color := ColorToRGB(FSelectionColor) xor ColorToRGB(Color);
end;

procedure TCustomZoomDisplay.SetDrawSelectionMode(Value: TDrawSelectionMode);
begin
  if Value <> FDrawSelectionMode then
  begin
    DrawSelection;
    if FDrawSelectionMode = smPatInvert then
    begin
      FSelBrush.Free;
      FSelBrush := nil;
    end else
    if Value = smPatInvert then
    begin
      FSelBrush := TBrush.Create;
      UpdateSelectionBrush;
    end;
    FDrawSelectionMode := Value;
    DrawSelection;
    Invalidate;
  end;
end;

procedure TCustomZoomDisplay.DrawSelection;
var
  OldBrush: HBrush;
  R: TRect;
begin
  case FDrawSelectionMode of
    smInvert: InvertRect(WaveformCanvas.Handle, GetSelectionRect);
    smPatInvert:
	if (FBitmap <> nil) and FMonochromeBitmap then
	  InvertRect(WaveformCanvas.Handle, GetSelectionRect)
	else
	begin
	  OldBrush := SelectObject(WaveformCanvas.Handle, FSelBrush.Handle);
	  try
	    R := GetSelectionRect;
	    PatBlt(WaveformCanvas.Handle, R.Left, R.Top, R.Right-R.Left, R.Bottom-R.Top, PATINVERT);
	  finally
	    SelectObject(WaveformCanvas.Handle, OldBrush);
	  end;
	end;
    // else do nothing
  end;
end;

function TCustomZoomDisplay.GetSelectionRect: TRect;
begin
  IntersectRect(Result, Rect(Round(DeviceX(FSelectionStart)), 0, Round(DeviceX(FSelectionEnd))+1, Height), Rect(0, 0, FViewport.W+1, FViewport.H+1));
end;

procedure TCustomZoomDisplay.DrawScales;
begin
  with WaveformCanvas, FViewport do
  begin
    if FHorzScale.Visible then
      FillRect(Rect(-X, H+1, ClientWidth-X, ClientHeight));
    if FVertScale.Visible then
      FillRect(Rect(-X, 0, 0, H+1));
    FVertScale.Draw(WaveformCanvas);
    FHorzScale.Draw(WaveformCanvas);
  end;
end;

procedure TCustomZoomDisplay.Paint;
begin
{$IFDEF Profile}
  begin
    Profile.Mark(1, True);
{$ENDIF}
  WaveformPaint;
{$IFDEF Profile}
    Profile.Mark(1, False);
  end;
{$ENDIF}
  if csDesigning in ComponentState then
  begin
    with Canvas do
    begin
      Brush.Style := bsClear;
      Pen.Style := psDash;
      Pen.Color := clWhite;
      Pen.Mode := pmXor;
      SetViewportOrgEx(Handle, 0, 0, nil);
      Rectangle(0, 0, Width, Height);
    end;
  end;
end;

function TCustomZoomDisplay.GetWaveformCanvas: TCanvas;
begin
  Result := FMetafileCanvas;
  if Result = nil then
  if (FBitmap <> nil)
    then Result := FBitmap.Canvas
    else Result := inherited Canvas;
{$IFDEF Profile}
Profile.Mark(2, True);
{$ENDIF}
  with FViewport do SetViewportOrgEx(Result.Handle, X, Y, nil);
{$IFDEF Profile}
Profile.Mark(2, False);
{$ENDIF}
end;

procedure TCustomZoomDisplay.DoWaveformPaint;
begin
  if Assigned(FOnPaint) then FOnPaint(Self);
end;

procedure TCustomZoomDisplay.WaveformPaint;
var
  R: TRect;
  C: TCanvas;
begin
  if FInWaveformPaint then Exit;
  FInWaveformPaint := True;
  try
    C := WaveformCanvas;
{$IFNDEF Ver90}
    C.Lock;
    try
{$ENDIF}
      C.Pen := FPen;
      C.Brush := Brush;
      R := Rect(0, 0, FViewport.W+1, FViewport.H+1);
  {$IFDEF Ver90}
      if (FBitmap <> nil) and FBitmap.Monochrome then
      begin
        C.Brush.Color := clWhite;    // otherwise bitmap background is dithered
        Canvas.Brush.Color := Color; // otherwise bitmap background is white
      end;
  {$ELSE}
      if FRestoreColors then
      begin
        if (FBitmap <> nil) and FBitmap.Monochrome then
          SetBitmapColors(FBitmap, [Color, FPen.Color], 0);// destroys ViewportOrg
        with FViewport do SetViewportOrgEx(C.Handle, X, Y, nil);
        FRestoreColors := False;
      end;
  {$ENDIF}
      if (FBitmap = nil) or FBitmapInvalid then
      begin
        C.FillRect(R);
        DrawScales;
        if not (csDesigning in ComponentState) then
        begin
          DoWaveformPaint;
          DrawSelection;
          DrawMarkerPositions;
          DrawHLSample;
        end;
        FBitmapInvalid := False;
      end;
      if FBitmap <> nil then
      begin
        SetViewportOrgEx(FBitmap.Canvas.Handle, 0, 0, nil);
        Canvas.Draw(0, 0, FBitmap);
      end;
{$IFNDEF Ver90}
    finally
      C.Unlock;
    end;
{$ENDIF}
  finally
    FInWaveformPaint := False;
  end;
end;

var
  ZoomBox: TRect;

procedure TCustomZoomDisplay.DrawZoomBox;
var
  L, T, R, B: Integer;
begin
// Rectangle doesn't include the right column and bottom row of pixels.
  with ZoomBox do
  begin
    L := Left;	R := Right;
    T := Top;	B := Bottom;
    if R >= L then Inc(R) else Inc(L);
    if B >= T then Inc(B) else Inc(T);
    Canvas.Rectangle(L, T, R, B);
  end;
end;

procedure TCustomZoomDisplay.BeginZoom(X, Y: Integer);
var
  R: TRect;
begin
  if FZooming then Exit;
  FZooming := True;
  Update;
  R := Viewport;
  if X < R.Left   then X := R.Left else
  if X > R.Right  then X := R.Right;
  if Y < R.Top    then Y := R.Top else
  if Y > R.Bottom then Y := R.Bottom;
  ZoomBox.Left	 := X;
  ZoomBox.Top	 := Y;
  ZoomBox.Right  := X;
  ZoomBox.Bottom := Y;
  MapWindowPoints(Handle, hWnd_Desktop, R, 2);
  ClipCursor(@R);
  MouseCapture := True;
  with Canvas do
  begin
    Brush.Style := bsClear;	// sets Brush.Color to clWhite
    Pen.Style := psDot;
    Pen.Color := clWhite;
    Pen.Mode := pmXor;
  end;
  DrawZoomBox;
end;

procedure TCustomZoomDisplay.ContinueZoom(X, Y: Integer);
begin
  DrawZoomBox;
  ZoomBox.Right  := X;
  ZoomBox.Bottom := Y;
  DrawZoomBox;
end;

procedure TCustomZoomDisplay.EndZoom;
begin
  DrawZoomBox;
  Dec(ZoomBox.Right);
  Dec(ZoomBox.Bottom);
  ClipCursor(nil);
  FZooming := False;
  MouseCapture := False;
  with Canvas do
  begin
  { BeginZoom sets Brush.Style to bsClear, which has the side
    effect to change Brush.Color to clWhite; hence I replaced
    "Brush.Style := bsSolid" by	"Brush.Color := Color" with
    the side effect of changing Brush.Style to bsSolid. }
    Brush.Color := Color;	// sets Brush.Style to bsSolid
    Pen := FPen;
  end;
  with FViewport do OffsetRect(ZoomBox, -X, -Y);
  RealizeZoom;
end;

procedure TCustomZoomDisplay.GetZoomWindow(var AWindow: TRect);
var
  dX, dY, Temp: Integer;
  dZoomBoxX: integer;
  dZoomBoxY: integer;
begin
  with ZoomBox do
  begin
    if Right < Left then
    begin
      Temp := Left;
      Left := Right;
      Right := Temp;
    end else
    if Right = Left then Right := Left + 1;
    if Bottom < Top then
    begin
      Temp := Top;
      Top := Bottom;
      Bottom := Temp;
    end else
    if Bottom = Top then Bottom := Top + 1;
  end;
  with AWindow do
  if (FZoomMode = zmZoomOut)
  or (FZoomMode = zmZoomIn) and FAlternateZoomModes then
  begin
    dX := FWindow.Right - FWindow.Left;
    dY := FWindow.Bottom - FWindow.Top;
    dZoomBoxX := ZoomBox.Right-ZoomBox.Left;
    dZoomBoxY := ZoomBox.Bottom-ZoomBox.Top;
    try
      Left   := FWindow.Left - IMulDiv(ZoomBox.Left, dX, dZoomBoxX);
      Right  := FWindow.Left + IMulDiv(FViewport.W,  dX, dZoomBoxX);
    except
      on EIntError do
	begin
	  AWindow.Left := FWorld.Left;
	  AWindow.Right := FWorld.Right;
	end;
    end;
    try
      Top    := FWindow.Top  - IMulDiv(ZoomBox.Top,  dY, dZoomBoxY);
      Bottom := FWindow.Top  + IMulDiv(FViewport.H,  dY, dZoomBoxY);
    except
      on EIntError do
	begin
	  AWindow.Top := FWorld.Top;
	  AWindow.Bottom := FWorld.Bottom;
	end;
    end;
  end else
  begin
    Left   := Round(WorldX(ZoomBox.Left));
    Top    := Round(WorldY(ZoomBox.Top));
    Right  := Round(WorldX(ZoomBox.Right));
    Bottom := Round(WorldY(ZoomBox.Bottom));
  end;
  BoundWindow(AWindow);
end;

procedure TCustomZoomDisplay.RealizeZoom;
var
  ZoomWindow: TRect;
begin
  GetZoomWindow(ZoomWindow);
  if not EqualRect(FWindow, ZoomWindow) then SetWindow(ZoomWindow);
end;

procedure TCustomZoomDisplay.WMSize(var Message: TWMSize);
begin
  inherited;
  if FBitmap <> nil then
//  if not (csDesigning in ComponentState) then
  begin
    if FBitmap.Width <> Message.Width then
    begin
      // In design mode: FBitmap.Width = 0 -> EOutOfResources
      if Message.Width = 0
        then FBitmap.Width := 1
        else FBitmap.Width := Message.Width;
      FBitmapInvalid := True;
      FRestoreColors := True;
    end;
    if FBitmap.Height <> Message.Height then
    begin
      // In design mode: FBitmap.Height = 0 -> EOutOfResources
      if Message.Height = 0
        then FBitmap.Height := 1
        else FBitmap.Height := Message.Height;
      FBitmapInvalid := True;
      FRestoreColors := True;
    end;
  end;
  InvalidateWorldWindow;
end;

procedure TCustomZoomDisplay.WMHScroll(var Message: TWMHScroll);
begin
  if Message.ScrollBar = 0 then HScroll(Message) else inherited;
end;

procedure TCustomZoomDisplay.WMVScroll(var Message: TWMVScroll);
begin
  if Message.ScrollBar = 0 then VScroll(Message) else inherited;
end;

procedure TCustomZoomDisplay.ScrollBy(Dx, Dy: Integer);
begin
  ScrollTo(FWindow.Left + Dx, FWindow.Bottom + Dy);
end;

procedure TCustomZoomDisplay.ScrollTo(X, Y: Integer);
begin
  X := iMax(FWorld.Left, iMin(X, FWorld.Right-(FWindow.Right-FWindow.Left)));
  Y := iMax(FWorld.Bottom, iMin(Y, FWorld.Top-(FWindow.Top-FWindow.Bottom)));
  if (X <> FXPos) or (Y <> FYPos) then
  begin
    FXPos := X;
    FYPos := Y;
    InvalidateWorldWindow;
  end;
end;

procedure TCustomZoomDisplay.HScroll(var Message: TWMScroll);
var
  ScrollInfo: TScrollInfo;
begin
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.fMask := HorzScrollBar.GetMask(SIF_ALL);
  GetScrollInfo(Handle, SB_HORZ, ScrollInfo);
  with Message do
    case ScrollCode of
      SB_LINEUP: ScrollBy(-FXIncrement, 0);
      SB_LINEDOWN: ScrollBy(FXIncrement, 0);
      SB_PAGEUP: ScrollBy(-WindowWidth, 0);
      SB_PAGEDOWN: ScrollBy(WindowWidth, 0);
      SB_THUMBPOSITION: ScrollTo(ScrollInfo.nPos, FYPos);
      SB_THUMBTRACK:
	with FHorzScrollBar do
        if Tracking then
          ScrollTo(ScrollInfo.nTrackPos, FYPos)
        else
        begin
	  ScrollInfo.nPos := ScrollInfo.nTrackPos;
	  SetScrollInfo(Handle, SB_HORZ, ScrollInfo, False);
	end;
      SB_TOP: ScrollTo(FWorld.Left, FYPos);
      SB_BOTTOM: ScrollTo(FWorld.Right-WindowWidth, FYPos);
      SB_ENDSCROLL: ;
    end;
end;

procedure TCustomZoomDisplay.VScroll(var Message: TWMScroll);
var
  ScrollInfo: TScrollInfo;
begin
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.fMask := VertScrollBar.GetMask(SIF_ALL);
  GetScrollInfo(Handle, SB_VERT, ScrollInfo);
  with Message do
    case ScrollCode of
      SB_LINEUP: ScrollBy(0, FYIncrement);
      SB_LINEDOWN: ScrollBy(0, -FYIncrement);
      SB_PAGEUP: ScrollBy(0, WindowHeight);
      SB_PAGEDOWN: ScrollBy(0, -WindowHeight);
      SB_THUMBPOSITION: ScrollTo(FXPos, 1-ScrollInfo.nPos-WindowHeight);
      SB_THUMBTRACK:
	with FVertScrollBar do
        if Tracking then
            ScrollTo(FXPos, 1-ScrollInfo.nTrackPos-WindowHeight)
        else
	begin
	  ScrollInfo.nPos := ScrollInfo.nTrackPos;
	  SetScrollInfo(Handle, SB_VERT, ScrollInfo, False);
	end;
      SB_TOP: ScrollTo(FXPos, FWorld.Top-WindowHeight);
      SB_BOTTOM: ScrollTo(FXPos, FWorld.Bottom);
      SB_ENDSCROLL: ;
    end;
end;

procedure TCustomZoomDisplay.UpdateIncrements;
begin
  FXIncrement := iMax(WindowWidth div FHorzScrollBar.PageFraction, 1);
  FYIncrement := iMax(WindowHeight div FVertScrollBar.PageFraction, 1);
end;

procedure TCustomZoomDisplay.IterateMarkers(Proc: TMarkerProc; Param: DWord);
var
  i: Integer;
  M: TMarker;
begin
  for i := 0 to ControlCount-1 do
  begin
    M := TMarker(Controls[i]);
    if M is TMarker then Proc(M, Param);
  end;
end;

procedure TCustomZoomDisplay.UpdateMarkers;
var
  i: Integer;
  M: TMarker;
begin
  for i := 0 to ControlCount-1 do
  begin
    M := TMarker(Controls[i]);
    if M is TMarker then M.UpdatePos;
  end;
end;

procedure TCustomZoomDisplay.DeleteMarkersMarking(Marked: TObject);
var
  i: Integer;
  M: TMarker;
begin
  i := 0;
  while i < ControlCount do
  begin
    M := TMarker(Controls[i]);
    if (M is TMarker) and (M.Marked = Marked)
      then M.Free
      else Inc(i);
  end;
end;

procedure TCustomZoomDisplay.SetMinVertScaleWidth(Value: Integer);
begin
  if Value <> FMinVertScaleWidth then
  begin
    FMinVertScaleWidth := Value;
    if FVertScale.Visible then InvalidateWorldWindow;
  end;
end;

procedure TCustomZoomDisplay.UpdateHorzScale;
var
  X1, X2: Extended;
begin
  X1 := FWindow.Left;
  X2 := FWindow.Right;
  if Assigned(FOnGetHorzScalePoint) then
  begin
    FOnGetHorzScalePoint(Self, X1);
    FOnGetHorzScalePoint(Self, X2);
  end;
  FHorzScale.SetSection(X1, X2);
end;

procedure TCustomZoomDisplay.UpdateVertScale;
var
  Y1, Y2: Extended;
begin
  Y1 := FWindow.Bottom;
  Y2 := FWindow.Top;
  if Assigned(FOnGetVertScalePoint) then
  begin
    FOnGetVertScalePoint(Self, Y1);
    FOnGetVertScalePoint(Self, Y2);
  end;
  FVertScale.SetSection(Y1, Y2);
  if FVertScale.Width > FViewport.X then UpdateViewport;
end;

procedure TCustomZoomDisplay.UpdateScales;
begin
  if FHorzScale.Visible then UpdateHorzScale;
  if FVertScale.Visible then UpdateVertScale;
end;

procedure TCustomZoomDisplay.UpdateViewport;
begin
  if csReading in ComponentState then Exit;
  with FViewport do
  begin
    if HorzScale.Visible then dH := HorzScale.Height+1 else dH := 1;
    if VertScale.Visible then
    begin
      dW := VertScale.Width+3;
      if dW < FMinVertScaleWidth+1 then dW := FMinVertScaleWidth+1;
    end else dW := 1;
    X := dW-1;
    SetViewport(X, Y, ClientWidth-dW+1, ClientHeight-dH+1);
  end;
  if not (csDestroying in ComponentState) then UpdateMarkers;
end;

function TCustomZoomDisplay.UpdateWorldWindow: Boolean;
begin
  Result := FWindowInvalid;
  if Result then
  begin
    MoveWindow(FXPos, FYPos);
    UpdateIncrements;
    UpdateScrollBars;
{$IFDEF Profile}
Profile.Mark(4, True);
{$ENDIF}
    UpdateViewport;
{$IFDEF Profile}
Profile.Mark(4, False);
{$ENDIF}
{$IFDEF Profile}
Profile.Mark(3, True);
{$ENDIF}
    UpdateScales;
{$IFDEF Profile}
Profile.Mark(3, False);
{$ENDIF}
    UpdateHighlightedX;
    if not EqualRect(FPrevWindow, FWindow) then
    begin
      FPrevWindow := FWindow;
      WindowChange;
    end;
    FBitmapInvalid := True;
    FWindowInvalid := False;
  end;
end;

function TCustomZoomDisplay.WorldX(X: Integer): TWorldCoord;
begin
  Result := FWindow.Left + (FWindow.Right-FWindow.Left) * (X / FViewport.W);
end;

function TCustomZoomDisplay.WorldY(Y: Integer): TWorldCoord;
begin
  Result := FWindow.Top + (FWindow.Bottom-FWindow.Top) * (Y / FViewport.H);
end;

procedure TCustomZoomDisplay.WorldPosChanged(MouseX, MouseY: Integer);
begin
  if not FMouseOver then Exit;
  Dec(MouseX, FViewport.X);
  Dec(MouseY, FViewport.Y);
  FMouseX := MouseX;
  FMouseY := MouseY;
  if Assigned(FOnWorldPosChange) then
    FOnWorldPosChange(Self, WorldX(MouseX), WorldY(MouseY));
end;

procedure TCustomZoomDisplay.WindowChange;
begin
  if Assigned(FOnWindowChange) then FOnWindowChange(Self);
  WorldPosChanged(FMouseX, FMouseY);
  if (MinWindowWidth <> MaxWindowWidth)
  or (MinWindowHeight <> MaxWindowHeight) then
  begin
    if Assigned(FOnZoomMax) then
    if (WindowWidth = MinWindowWidth) and (WindowHeight = MinWindowHeight) then
      FOnZoomMax(Self);
    if Assigned(FOnZoomMin) then
    if (WindowWidth = MaxWindowWidth) and (WindowHeight = MaxWindowHeight) then
      FOnZoomMin(Self);
  end;
end;

procedure TCustomZoomDisplay.SaveToFile(FileName: string);
var
  M: TMetafile;
  WasBuffered: Boolean;
begin
  WasBuffered := Buffered;
  try
    case ImageFileFormat of
      imBMP: Bitmap.SaveToFile(FileName);
      imEMF, imWMF:
	begin
	  M := TMetafile.Create;
	  try
	    Buffered := False;
	    M.Enhanced := ImageFileFormat = imEMF;
	    M.Width := FViewport.W;
	    M.Height := FViewport.H;
	    FMetaFileCanvas := TMetaFileCanvas.Create(M, Canvas.Handle);
	    FDoDrawMarkerPositions := False;
	    try
	      Paint;
	    finally
	      FDoDrawMarkerPositions := True;
	      FMetafileCanvas.Free;
	      FMetafileCanvas := nil;
	    end;
	    M.SaveToFile(FileName);
	  finally
	    M.Free;
	  end;
	end;
    end;
  finally
    Buffered := WasBuffered;
  end;
end;

function BytesPerScanline(PixelsPerScanline, BitsPerPixel, Alignment: Longint): Longint;
begin
  Dec(Alignment);
  Result := ((PixelsPerScanline * BitsPerPixel) + Alignment) and not Alignment;
  Result := Result div 8;
end;

procedure TCustomZoomDisplay.VerticalLine(DC: hDC; X, Ymin, Ymax: Integer);
var
  P: TPoint;
begin
  if X < 0 then Exit;
  Ymin := iMin(FViewport.H, Ymin);
  if Ymin < 0 then Exit;
  Ymax := iMax(0, Ymax);
  if Ymax > FViewport.H then Exit;
  Windows.MoveToEx(DC, X, Ymin, @P);
  Windows.LineTo(DC, X, Ymax-1);  // -1 to compensate for the fact that LineTo doesn't draw the last point
end;

procedure TCustomZoomDisplay.PlotSamples(
  Sample: Pointer;
  Amplitude: TAmpFunc);
var
  DC: hDC;
  P: Pointer;
  AmpMin: TWorldCoord;
  AmpMax: TWorldCoord;
  PixelHeight: Double;
  X, Y, W, H: Integer;
  Ymin: Integer;
  Ymax: Integer;
  YminPrev: Integer;
  YmaxPrev: Integer;
  WX: Integer;
  WY: TWorldCoord;
  WX1: Integer;
  WX2: Integer;
  WY2: Integer;
  WXmax: Integer;
{$IFNDEF VER90}
  MonoBmp: Boolean;
  Bits: PByteArray;
  ScanLine0: Integer;
  ScanLineW: Integer;

  procedure VerticalLine(DC: hDC; X, Ymin, Ymax: Integer);
  var
    Y: Integer;
    Mask: DWORD;
    ScanLineY: PByteArray;
  begin
    if not MonoBmp then
    begin
      Self.VerticalLine(DC, X, Ymin, Ymax);
      Exit;
    end;
    if X < 0 then Exit;
    Y := iMin(H, Ymin);
    if Y < 0 then Exit;
    Inc(X, FViewport.X);
    Mask := $80 shr (X and 7);
    ScanLineY := PByteArray(Integer(Bits)+ScanLineW*(ScanLine0-Y)); // PByteArray(Bitmap.ScanLine[Y]);
    for Y := Y downto iMax(0, Ymax) do
    begin
      ScanLineY^[X shr 3] := ScanLineY^[X shr 3] or Mask;
      Inc(PByte(ScanLineY), ScanLineW);
    end;
  end;
{$ENDIF}
  procedure GetMinMax(Start, Stop: Integer);
  var
    Amp: TWorldCoord;
    i: Integer;
    P: Pointer;
  begin
    P := Sample;
    Amp := WY;
    for i := Start to Stop do
    begin
       Amp := Amplitude(P);
       if Amp < AmpMin then AmpMin := Amp;
       if Amp > AmpMax then AmpMax := Amp;
    end;
    WY := Amp;
    Sample := P;
  end;

begin
{$IFNDEF VER90}
  MonoBmp := FMonochromeBitmap and (FBitmap <> nil);
  if MonoBmp then with FBitmap do
  begin
    ScanLineW := BytesPerScanline(Width, 1, 32);
    ScanLine0 := Height-1;
    Bits := ScanLine[ScanLine0];
  end;
{$ENDIF}
  DC := WaveformCanvas.Handle;
  WX1 := Window.Left;
  WX2 := Window.Right;
  W := FViewport.W;
  P := Sample;
  if W > WX2-WX1 then
  begin
    MoveTo(WX1, Amplitude(P));
    for WX := Succ(WX1) to WX2 do DoLineTo(DC, WX, Amplitude(P));
  end else
  begin
    H := FViewport.H;
    WY2 := Window.Top;
    PixelHeight := H/(Window.Bottom-WY2);
    WX := WX1;
    WY := Amplitude(P);
    YMinPrev := Round((WY-WY2)*PixelHeight);
    YmaxPrev := YminPrev;
    for X := 0 to W do
    begin
      P := Sample;
      Y := Trunc(((WY+Amplitude(P))/2-WY2)*PixelHeight);
      if FPaintMode = spHollow then
      begin
        AmpMin := FWorld.Top;
        AmpMax := FWorld.Bottom;
      end else // PaintMode = spSolid
      begin
        AmpMin := 0;
        AmpMax := 0;
      end;
      WXmax := iMin(WX2, WX1 + iMulDiv(X+1, WX2-WX1, W+1));
      GetMinMax(WX, WXmax);
      WX := WXmax+1;
      Ymin := Round((AmpMin-WY2)*PixelHeight);
      Ymax := Round((AmpMax-WY2)*PixelHeight);
      if Ymin < YmaxPrev-1 then
      begin
	Ymin := iMax(Ymin, Y);
	YmaxPrev := iMin(YmaxPrev, (Ymin+1));
      end else
      if YminPrev < Ymax-1 then
      begin
	YminPrev := iMax(YminPrev, Y);
	Ymax := iMin(Ymax, YminPrev+1);
      end;
      VerticalLine(DC, X-1, YminPrev, YmaxPrev);
      if WXmax = WX2 then
      begin
	VerticalLine(DC, X, Ymin, Ymax);
	exit;
      end;
      YminPrev := Ymin;
      YmaxPrev := Ymax;
    end;
  end;
end;

procedure TCustomZoomDisplay.PlotByteArray(
  const Samples: array of Byte;
  Offset: Longint;
  Channels: Integer;
  AChannel: Integer);
type
  TSample = Byte;
  PSample = ^TSample;
var
  DC: hDC;
  AmpMin: Integer;
  AmpMax: Integer;
  X, Y, W, H: Integer;
  Ymin: Integer;
  Ymax: Integer;
  YminPrev: Integer;
  YmaxPrev: Integer;
  WH: Integer;
  WX: Integer;
  WY: Integer;
  WX1: Integer;
  WX2: Integer;
  WY2: Integer;
  WXmax: Integer;
  WXBufEnd: Integer;
  Sample: PSample;
  BufRem: Integer;
  SampleBlockCount: Integer;
  InClipRect: Boolean;
{$IFNDEF VER90}
  MonoBmp: Boolean;
  Bits: PByteArray;
  ScanLine0: Integer;
  ScanLineW: Integer;

  procedure VerticalLine(DC: hDC; X, Ymin, Ymax: Integer);
  var
    Y: Integer;
    Mask: DWORD;
    ScanLineY: PByteArray;
  begin
    if not InClipRect then Exit;
    if not MonoBmp then
    begin
      Self.VerticalLine(DC, X, Ymin, Ymax);
      Exit;
    end;
    if X < 0 then Exit;
    Y := iMin(H, Ymin);
    if Y < 0 then Exit;
    Inc(X, FViewport.X);
    Mask := $80 shr (X and 7);
    ScanLineY := PByteArray(Integer(Bits)+ScanLineW*(ScanLine0-Y)); // PByteArray(Bitmap.ScanLine[Y]);
    for Y := Y downto iMax(0, Ymax) do
    begin
      ScanLineY^[X shr 3] := ScanLineY^[X shr 3] or Mask;
      Inc(PByte(ScanLineY), ScanLineW);
    end;
  end;
{$ENDIF}

  procedure GetMinMax(First, Last: Integer);	assembler;
  {
  var
    i: Integer;
  begin
    if InClipRect then
    for i := First to Last do
    begin
       WY := Sample^;
       if WY < AmpMin then AmpMin := WY;
       if WY > AmpMax then AmpMax := WY;
       Inc(Sample, Channels);
    end else
    begin
      Inc(Sample, Channels*(Last-First));
      WY := Sample^;
      Inc(Sample, Channels);
    end;
    WX := Last+1;
  end;
    // Min in dl, Max in dh
  }
  asm
	mov	ecx, Last
        inc     ecx
        mov     WX, ecx
	sub	ecx, First
	jle	@end
	push	esi
	push	edi
	mov	esi, Sample
	mov	edi, Channels
        test    InClipRect, 1
        jz      @2
	mov	dl, BYTE PTR AmpMin
	mov	dh, BYTE PTR AmpMax
  @GetNext:
	mov	al, BYTE PTR [esi]
	add	esi, edi
	cmp	dl, al
	jna	@GetMax
	mov	dl, al
  @GetMax:
	cmp	dh, al
	jna	@1
	dec	ecx
	jnz	@GetNext
	jmp	@exit
  @1:	mov	dh, al
	dec     ecx
	jnz	@GetNext
        jmp     @exit
  @2:   mov     eax, ecx
        dec     eax
        mul     edi
        add     esi, eax
	mov	al, BYTE PTR [esi]
        add     esi, edi
  @exit:
	mov	Sample, esi
	movzx	eax, al
	mov	WY, eax
	movzx	eax, dl
	mov	AmpMin, eax
	movzx	eax, dh
	mov	AmpMax, eax
	pop	edi
	pop	esi
  @end:
  end;

var
  R: TRect;
begin
{$IFNDEF VER90}
  MonoBmp := FMonochromeBitmap and (FBitmap <> nil);
  if MonoBmp then
  with FBitmap do
  begin
    ScanLineW := BytesPerScanline(Width, 1, 32);
    ScanLine0 := Height-1;
    Bits := ScanLine[ScanLine0];
  end;
{$ENDIF}
  DC := WaveformCanvas.Handle;
  WX1 := Window.Left;
  WX2 := Window.Right;
  Sample := PSample(@Samples);
  SampleBlockCount := (High(Samples)-Low(Samples)+1) div Channels;
  if SampleBlockCount < WX2-WX1+1 then Error(SPlotArray);
  BufRem := iMin(WX2-WX1+1, SampleBlockCount-Offset);
  Inc(Sample, Offset * Channels + AChannel);
  WX := WX1;
  W := FViewport.W;
  if W > WX2-WX1 then
  begin
    MoveTo(WX, Sample^);
    WXmax := WX + Pred(BufRem);
    for WX := Succ(WX) to WXmax do
    begin
      Inc(Sample, Channels);
      DoLineTo(DC, WX, Sample^);
    end;
    Sample := PSample(@Samples);
    Inc(Sample, AChannel);
    BufRem := WX2-WX1+1-BufRem;
    for WX := Succ(WXmax) to WXmax + Pred(BufRem) do
    begin
      DoLineTo(DC, WX, Sample^);
      Inc(Sample, Channels);
    end;
  end else
  begin
    H := FViewport.H;
    WY2 := Window.Top;
    WH := Window.Bottom-WY2;
    WY := Sample^;
    YMinPrev := IMulDiv(WY-WY2, H, WH);
    YmaxPrev := YminPrev;
    R := WaveformCanvas.ClipRect;
    for X := 0 to iMin(R.Right, W) do
    begin
      InClipRect := not ((X < R.Left) or (X > R.Right));
      Y := Trunc(((WY+Sample^)/2-WY2)*H/WH);
      if FPaintMode = spHollow then
      begin
        AmpMin := High(TSample);
        AmpMax := Low(TSample);
      end else // PaintMode = spSolid
      begin
        AmpMin := $80;
        AmpMax := $80;
      end;
      WXmax := iMin(WX2, WX1 + iMulDiv(X+1, WX2-WX1, W+1));
      WXBufEnd := WX+BufRem-1;
      if WXmax > WXBufEnd then
      begin
	GetMinMax(WX, WXBufEnd);
	WX := WXBufEnd+1;
	BufRem := iMin(WX2-WX+1, SampleBlockCount);
	Sample := PSample(@Samples);
	Inc(Sample, AChannel);
      end;
      Dec(BufRem, WXmax-WX+1);
      GetMinMax(WX, WXmax);
      WX := WXmax+1;
      Ymin := IMulDiv(AmpMin-WY2, H, WH);
      Ymax := IMulDiv(AmpMax-WY2, H, WH);
      if Ymin < YmaxPrev-1 then
      begin
	Ymin := iMax(Ymin, Y);
	YmaxPrev := iMin(YmaxPrev, (Ymin+1));
      end else
      if YminPrev < Ymax-1 then
      begin
	YminPrev := iMax(YminPrev, Y);
	Ymax := iMin(Ymax, YminPrev+1);
      end;
      VerticalLine(DC, X-1, YminPrev, YmaxPrev);
      if WXmax = WX2 then
      begin
	VerticalLine(DC, X, Ymin, Ymax);
	exit;
      end;
      YminPrev := Ymin;
      YmaxPrev := Ymax;
    end;
  end;
end;

procedure TCustomZoomDisplay.PlotSmallintArray(
  const Samples: array of Smallint;
  Offset: Longint;
  Channels: Integer;
  AChannel: Integer);
{ Die Idee hinter der Routine fr den Fall, da das
  Weltfenster breiter ist als das Bildschirmfenster,
  ist folgende:
  Wenn mehr als eine Welt-X-Koordinate (Wx) auf
  eine Bildschirm-X-Koordinate (X) entfllt, gehe
  man so vor:
1. Aufeinanderfolgende Samples, die sich in der
   horizontalen Bildschirmkoordinate unterscheiden,
   werden durch eine Linie verbunden;
2. Haben mehrere Samples dieselbe horizontale
   Bildschirmkoordinate, so werden Minimum und
   Maximum ihrer Amplituden festgestellt und
   durch eine senkrechte Linie dargestellt.
3. Das legt eine Verfeinerung nahe: Die Hlfte
   der Linie von 1. jeweils der entsprechenden
   senkrechten aus 2. zuschlagen. }
type
  TSample = Smallint;
  PSample = ^TSample;
var
  DC: hDC;
  AmpMin: Integer;
  AmpMax: Integer;
  X, Y, W, H: Integer;
  Ymin: Integer;
  Ymax: Integer;
  YminPrev: Integer;
  YmaxPrev: Integer;
  WH: Integer;
  WX: Integer;
  WY: Integer;
  WX1: Integer;
  WX2: Integer;
  WY2: Integer;
  WXmax: Integer;
  WXBufEnd: Integer;
  Sample: PSample;
  BufRem: Integer;
  SampleBlockCount: Integer;
  InClipRect: Boolean;
{$IFNDEF VER90}
  MonoBmp: Boolean;
  Bits: PByteArray;
  ScanLine0: Integer;
  ScanLineW: Integer;

  procedure VerticalLine(DC: hDC; X, Ymin, Ymax: Integer);
  var
    Y: Integer;
    Mask: DWORD;
    ScanLineY: PByteArray;
  begin
    if not InClipRect then Exit;
    if not MonoBmp then
    begin
      Self.VerticalLine(DC, X, Ymin, Ymax);
      Exit;
    end;
    if X < 0 then Exit;
    Y := iMin(H, Ymin);
    if Y < 0 then Exit;
    Inc(X, FViewport.X);
    Mask := $80 shr (X and 7);
    ScanLineY := PByteArray(Integer(Bits)+ScanLineW*(ScanLine0-Y)); // PByteArray(Bitmap.ScanLine[Y]);
    for Y := Y downto iMax(0, Ymax) do
    begin
      ScanLineY^[X shr 3] := ScanLineY^[X shr 3] or Mask;
      Inc(PByte(ScanLineY), ScanLineW);
    end;
  end;
{$ENDIF}
  procedure GetMinMax(First, Last: Integer);	assembler;
  {
  var
    i: Integer;
  begin
    if InClipRect then
    for i := First to Last do
    begin
       WY := Sample^;
       if WY < AmpMin then AmpMin := WY;
       if WY > AmpMax then AmpMax := WY;
       Inc(Sample, Channels);
    end else
    begin
      Inc(Sample, Channels*(Last-First));
      WY := Sample^;
      Inc(Sample, Channels);
    end;
    WX := Last+1;
  end;
    // Min in bx, Max in dx
  }
  asm
	mov	ecx, Last
        inc     ecx
        mov     WX, ecx
	sub	ecx, First
	jle	@end
	push	ebx
	push	esi
	push	edi
	mov	esi, Sample
	mov	edi, Channels
	sal	edi, 1	// edi := Channels * SizeOf(Sample^)
        test    InClipRect, 1
        jz      @2
	mov	bx, WORD PTR AmpMin
	mov	dx, WORD PTR AmpMax
  @GetNext:
	mov	ax, WORD PTR [esi]
	add	esi, edi
	cmp	bx, ax
	jng	@GetMax
	mov	bx, ax
  @GetMax:
	cmp	dx, ax
	jng	@1
	dec	ecx
	jnz	@GetNext
	jmp	@exit
  @1:	mov	dx, ax
	dec     ecx
	jnz	@GetNext
        jmp     @exit
  @2:   mov     eax, ecx
        dec     eax
        mul     edi
        add     esi, eax
	mov	ax, WORD PTR [esi]
        add     esi, edi
  @exit:
	mov	Sample, esi
	cwde
	mov	WY, eax
	movsx	eax, bx
	mov	AmpMin, eax
	movsx	eax, dx
	mov	AmpMax, eax
	pop	edi
	pop	esi
	pop	ebx
  @end:
  end;

var
  R: TRect;
begin
{$IFNDEF VER90}
  MonoBmp := FMonochromeBitmap and (FBitmap <> nil);
  if MonoBmp then
  with FBitmap do
  begin
    ScanLineW := BytesPerScanline(Width, 1, 32);
    ScanLine0 := Height-1;
    Bits := ScanLine[ScanLine0];
  end;
{$ENDIF}
  DC := WaveformCanvas.Handle;
  WX1 := Window.Left;
  WX2 := Window.Right;
  Sample := PSample(@Samples);
  SampleBlockCount := (High(Samples)-Low(Samples)+1) div Channels;
  if SampleBlockCount < WX2-WX1+1 then Error(SPlotArray);
  BufRem := iMin(WX2-WX1+1, SampleBlockCount-Offset);
  Inc(Sample, Offset * Channels + AChannel);
  WX := WX1;
  W := FViewport.W;
  if W > WX2-WX1 then
  begin
    MoveTo(WX, Sample^);
    WXmax := WX + Pred(BufRem);
    for WX := Succ(WX) to WXmax do
    begin
      Inc(Sample, Channels);
      DoLineTo(DC, WX, Sample^);
    end;
    Sample := PSample(@Samples);
    Inc(Sample, AChannel);
    BufRem := WX2-WX1+1-BufRem;
    for WX := Succ(WXmax) to WXmax + Pred(BufRem) do
    begin
      DoLineTo(DC, WX, Sample^);
      Inc(Sample, Channels);
    end;
  end else
  begin
    H := FViewport.H;
    WY2 := Window.Top;
    WH := Window.Bottom-WY2;
    WY := Sample^;
    YMinPrev := IMulDiv(WY-WY2, H, WH);
    YmaxPrev := YminPrev;
    R := WaveformCanvas.ClipRect;
{$IFDEF Profile}
    Profile.Mark(5, True);
{$ENDIF}
    for X := 0 to iMin(R.Right, W) do
    begin
      InClipRect := not ((X < R.Left) or (X > R.Right));
      Y := Trunc(((WY+Sample^)/2-WY2)*H/WH);
      if FPaintMode = spHollow then
      begin
        AmpMin := High(TSample);
        AmpMax := Low(TSample);
      end else // PaintMode = spSolid
      begin
        AmpMin := 0;
        AmpMax := 0;
      end;
      WXmax := iMin(WX2, WX1 + iMulDiv(X+1, WX2-WX1, W+1));
      WXBufEnd := WX+BufRem-1;
      if WXmax > WXBufEnd then
      begin
        GetMinMax(WX, WXBufEnd);
	BufRem := iMin(WX2-WX+1, SampleBlockCount);
	Sample := PSample(@Samples);
	Inc(Sample, AChannel);
      end;
      Dec(BufRem, WXmax-WX+1);
      GetMinMax(WX, WXmax);
      Ymin := IMulDiv(AmpMin-WY2, H, WH);
      Ymax := IMulDiv(AmpMax-WY2, H, WH);
      if Ymin < YmaxPrev-1 then
      begin
	Ymin := iMax(Ymin, Y);
	YmaxPrev := iMin(YmaxPrev, (Ymin+1));
      end else
      if YminPrev < Ymax-1 then
      begin
	YminPrev := iMax(YminPrev, Y);
	Ymax := iMin(Ymax, YminPrev+1);
      end;
      VerticalLine(DC, X-1, YminPrev, YmaxPrev);
      if WXmax = WX2 then
      begin
        VerticalLine(DC, X, Ymin, Ymax);
	Exit;
      end;
      YminPrev := Ymin;
      YmaxPrev := Ymax;
    end;
{$IFDEF Profile}
    Profile.Mark(5, False);
{$ENDIF}
  end;
end;

procedure TCustomZoomDisplay.Plot24BitSamples(
  const Samples: array of T24BitSample;
  Offset: Longint;
  Channels: Integer;
  AChannel: Integer);
type
  TSample = T24BitSample;
  PSample = ^TSample;
var
  DC: hDC;
  AmpMin: Integer;
  AmpMax: Integer;
  X, Y, W, H: Integer;
  Ymin: Integer;
  Ymax: Integer;
  YminPrev: Integer;
  YmaxPrev: Integer;
  WH: Integer;
  WX: Integer;
  WY: Integer;
  WX1: Integer;
  WX2: Integer;
  WY2: Integer;
  WXmax: Integer;
  WXBufEnd: Integer;
  Sample: PSample;
  BufRem: Integer;
  SampleBlockCount: Integer;
{$IFNDEF VER90}
  MonoBmp: Boolean;
  Bits: PByteArray;
  ScanLine0: Integer;
  ScanLineW: Integer;

  procedure VerticalLine(DC: hDC; X, Ymin, Ymax: Integer);
  var
    Y: Integer;
    Mask: DWORD;
    ScanLineY: PByteArray;
  begin
    if not MonoBmp then
    begin
      Self.VerticalLine(DC, X, Ymin, Ymax);
      Exit;
    end;
    if X < 0 then Exit;
    Y := iMin(H, Ymin);
    if Y < 0 then Exit;
    Inc(X, FViewport.X);
    Mask := $80 shr (X and 7);
    ScanLineY := PByteArray(Integer(Bits)+ScanLineW*(ScanLine0-Y)); // PByteArray(Bitmap.ScanLine[Y]);
    for Y := Y downto iMax(0, Ymax) do
    begin
      ScanLineY^[X shr 3] := ScanLineY^[X shr 3] or Mask;
      Inc(PByte(ScanLineY), ScanLineW);
    end;
  end;
{$ENDIF}

  function Amplitude(Sample: T24BitSample): Integer;
  asm
	mov	eax, DWord([Sample])
	shl	eax, 8	// sign extend 24 -> 32 bit, step 1
	sar	eax, 8	// step 2
  end;

  procedure GetMinMax(Start, Stop: Integer);	assembler;
  {
  var
    i: Integer;
  begin
    for i := Start to Stop do
    begin
       WY := Sample^;
       if WY < AmpMin then AmpMin := WY;
       if WY > AmpMax then AmpMax := WY;
       Inc(Sample, Channels);
    end;
  end;
    // Min in ebx, Max in edx
  }
  asm
	mov	ecx, Stop
	sub	ecx, Start
	jl	@end
	inc	ecx
	push	ebx
	push	esi
	push	edi
	mov	esi, Sample
	mov	eax, Channels
	mov	edx, TYPE TSample // edi := Channels * SizeOf(Sample^)
	mul	edx
	mov	edi, eax
	mov	ebx, AmpMin
	mov	edx, AmpMax
  @GetNext:
	mov	eax, [esi]
	shl	eax, 8	// sign extend 24 -> 32 bit, step 1
	sar	eax, 8	// step 2
	add	esi, edi
	cmp	ebx, eax
	jng	@GetMax
	mov	ebx, eax
  @GetMax:
	cmp	edx, eax
	jng	@1
	dec	ecx
	jnz	@GetNext
	jmp	@exit
  @1:	mov	edx, eax
	dec     ecx
	jnz	@GetNext
  @exit:
	mov	Sample, esi
	mov	WY, eax
	mov	AmpMin, ebx
	mov	AmpMax, edx
	pop	edi
	pop	esi
	pop	ebx
  @end:
  end;

begin
{$IFNDEF VER90}
  MonoBmp := FMonochromeBitmap and (FBitmap <> nil);
  if MonoBmp then
  with FBitmap do
  begin
    ScanLineW := BytesPerScanline(Width, 1, 32);
    ScanLine0 := Height-1;
    Bits := ScanLine[ScanLine0];
  end;
{$ENDIF}
  DC := WaveformCanvas.Handle;
  WX1 := Window.Left;
  WX2 := Window.Right;
  Sample := PSample(@Samples);
  SampleBlockCount := High(Samples)-Low(Samples)+1;
  if SampleBlockCount < WX2-WX1+1 then Error(SPlotArray);
  BufRem := iMin(WX2-WX1+1, SampleBlockCount-Offset);
  Inc(Sample, Offset * Channels + AChannel);
  WX := WX1;
  W := FViewport.W;
  if W > WX2-WX1 then
  begin
    MoveTo(WX, Amplitude(Sample^));
    WXmax := WX + Pred(BufRem);
    for WX := Succ(WX) to WXmax do
    begin
      Inc(Sample, Channels);
      DoLineTo(DC, WX, Amplitude(Sample^));
    end;
    Sample := PSample(@Samples);
    Inc(Sample, AChannel);
    BufRem := WX2-WX1+1-BufRem;
    for WX := Succ(WXmax) to WXmax + Pred(BufRem) do
    begin
      DoLineTo(DC, WX, Amplitude(Sample^));
      Inc(Sample, Channels);
    end;
  end else
  begin
    H := FViewport.H;
    WY2 := Window.Top;
    WH := Window.Bottom-WY2;
    WY := Amplitude(Sample^);
    YMinPrev := IMulDiv(WY-WY2, H, WH);
    YmaxPrev := YminPrev;
    for X := 0 to W do
    begin
      Y := Trunc(((WY+Amplitude(Sample^))/2-WY2)*H/WH);
      if FPaintMode = spHollow then
      begin
        AmpMin := High24BitSample;
        AmpMax := Low24BitSample;
      end else // PaintMode = spSolid
      begin
        AmpMin := 0;
        AmpMax := 0;
      end;
      WXmax := iMin(WX2, WX1 + iMulDiv(X+1, WX2-WX1, W+1));
      WXBufEnd := WX+BufRem-1;
      if WXmax > WXBufEnd then
      begin
	GetMinMax(WX, WXBufEnd);
	WX := WXBufEnd+1;
	BufRem := iMin(WX2-WX+1, SampleBlockCount);
	Sample := PSample(@Samples);
	Inc(Sample, AChannel);
      end;
      Dec(BufRem, WXmax-WX+1);
      GetMinMax(WX, WXmax);
      WX := WXmax+1;
      Ymin := IMulDiv(AmpMin-WY2, H, WH);
      Ymax := IMulDiv(AmpMax-WY2, H, WH);
      if Ymin < YmaxPrev-1 then
      begin
	Ymin := iMax(Ymin, Y);
	YmaxPrev := iMin(YmaxPrev, (Ymin+1));
      end else
      if YminPrev < Ymax-1 then
      begin
	YminPrev := iMax(YminPrev, Y);
	Ymax := iMin(Ymax, YminPrev+1);
      end;
      VerticalLine(DC, X-1, YminPrev, YmaxPrev);
      if WXmax = WX2 then
      begin
	VerticalLine(DC, X, Ymin, Ymax);
	exit;
      end;
      YminPrev := Ymin;
      YmaxPrev := Ymax;
    end;
  end;
end;

procedure TCustomZoomDisplay.PlotIntArray(
  const Samples: array of Integer;
  Offset: Longint;
  Channels: Integer;
  AChannel: Integer);
type
  TSample = Integer;
  PSample = ^TSample;
var
  DC: hDC;
  AmpMin: Integer;
  AmpMax: Integer;
  X, Y, W, H: Integer;
  Ymin: Integer;
  Ymax: Integer;
  YminPrev: Integer;
  YmaxPrev: Integer;
  WH: Integer;
  WX: Integer;
  WY: Integer;
  WX1: Integer;
  WX2: Integer;
  WY2: Integer;
  WXmax: Integer;
  WXBufEnd: Integer;
  Sample: PSample;
  BufRem: Integer;
  SampleBlockCount: Integer;
{$IFNDEF VER90}
  MonoBmp: Boolean;
  Bits: PByteArray;
  ScanLine0: Integer;
  ScanLineW: Integer;

  procedure VerticalLine(DC: hDC; X, Ymin, Ymax: Integer);
  var
    Y: Integer;
    Mask: DWORD;
    ScanLineY: PByteArray;
  begin
    if not MonoBmp then
    begin
      Self.VerticalLine(DC, X, Ymin, Ymax);
      Exit;
    end;
    if X < 0 then Exit;
    Y := iMin(H, Ymin);
    if Y < 0 then Exit;
    Inc(X, FViewport.X);
    Mask := $80 shr (X and 7);
    ScanLineY := PByteArray(Integer(Bits)+ScanLineW*(ScanLine0-Y)); // PByteArray(Bitmap.ScanLine[Y]);
    for Y := Y downto iMax(0, Ymax) do
    begin
      ScanLineY^[X shr 3] := ScanLineY^[X shr 3] or Mask;
      Inc(PByte(ScanLineY), ScanLineW);
    end;
  end;
{$ENDIF}
  procedure GetMinMax(Start, Stop: Integer);	assembler;
  {
  var
    i: Integer;
  begin
    for i := Start to Stop do
    begin
       WY := Sample^;
       if WY < AmpMin then AmpMin := WY;
       if WY > AmpMax then AmpMax := WY;
       Inc(Sample, Channels);
    end;
  end;
    // Min in ebx, Max in edx
  }
  asm
	mov	ecx, Stop
	sub	ecx, Start
	jl	@end
	inc	ecx
	push	ebx
	push	esi
	push	edi
	mov	esi, Sample
	mov	edi, Channels
	sal	edi, 2	// edi := Channels * SizeOf(Sample^)
	mov	ebx, AmpMin
	mov	edx, AmpMax
  @GetNext:
	mov	eax, [esi]
	add	esi, edi
	cmp	ebx, eax
	jng	@GetMax
	mov	ebx, eax
  @GetMax:
	cmp	edx, eax
	jng	@1
	dec	ecx
	jnz	@GetNext
	jmp	@exit
  @1:	mov	edx, eax
	dec     ecx
	jnz	@GetNext
  @exit:
	mov	Sample, esi
	mov	WY, eax
	mov	AmpMin, ebx
	mov	AmpMax, edx
	pop	edi
	pop	esi
	pop	ebx
  @end:
  end;

begin
{$IFNDEF VER90}
  MonoBmp := FMonochromeBitmap and (FBitmap <> nil);
  if MonoBmp then
  with FBitmap do
  begin
    ScanLineW := BytesPerScanline(Width, 1, 32);
    ScanLine0 := Height-1;
    Bits := ScanLine[ScanLine0];
  end;
{$ENDIF}
  DC := WaveformCanvas.Handle;
  WX1 := Window.Left;
  WX2 := Window.Right;
  Sample := PSample(@Samples);
  SampleBlockCount := High(Samples)-Low(Samples)+1;
  if SampleBlockCount < WX2-WX1+1 then Error(SPlotArray);
  BufRem := iMin(WX2-WX1+1, SampleBlockCount-Offset);
  Inc(Sample, Offset * Channels + AChannel);
  WX := WX1;
  W := FViewport.W;
  if W > WX2-WX1 then
  begin
    MoveTo(WX, Sample^);
    WXmax := WX + Pred(BufRem);
    for WX := Succ(WX) to WXmax do
    begin
      Inc(Sample, Channels);
      DoLineTo(DC, WX, Sample^);
    end;
    Sample := PSample(@Samples);
    Inc(Sample, AChannel);
    BufRem := WX2-WX1+1-BufRem;
    for WX := Succ(WXmax) to WXmax + Pred(BufRem) do
    begin
      DoLineTo(DC, WX, Sample^);
      Inc(Sample, Channels);
    end;
  end else
  begin
    H := FViewport.H;
    WY2 := Window.Top;
    WH := Window.Bottom-WY2;
    WY := Sample^;
    YMinPrev := IMulDiv(WY-WY2, H, WH);
    YmaxPrev := YminPrev;
    for X := 0 to W do
    begin
      Y := Trunc(((WY+Sample^)/2-WY2)*H/WH);
      if FPaintMode = spHollow then
      begin
        AmpMin := High(TSample);
        AmpMax := Low(TSample);
      end else // PaintMode = spSolid
      begin
        AmpMin := 0;
        AmpMax := 0;
      end;
      WXmax := iMin(WX2, WX1 + iMulDiv(X+1, WX2-WX1, W+1));
      WXBufEnd := WX+BufRem-1;
      if WXmax > WXBufEnd then
      begin
	GetMinMax(WX, WXBufEnd);
	WX := WXBufEnd+1;
	BufRem := iMin(WX2-WX+1, SampleBlockCount);
	Sample := PSample(@Samples);
	Inc(Sample, AChannel);
      end;
      Dec(BufRem, WXmax-WX+1);
      GetMinMax(WX, WXmax);
      WX := WXmax+1;
      Ymin := IMulDiv(AmpMin-WY2, H, WH);
      Ymax := IMulDiv(AmpMax-WY2, H, WH);
      if Ymin < YmaxPrev-1 then
      begin
	Ymin := iMax(Ymin, Y);
	YmaxPrev := iMin(YmaxPrev, (Ymin+1));
      end else
      if YminPrev < Ymax-1 then
      begin
	YminPrev := iMax(YminPrev, Y);
	Ymax := iMin(Ymax, YminPrev+1);
      end;
      VerticalLine(DC, X-1, YminPrev, YmaxPrev);
      if WXmax = WX2 then
      begin
	VerticalLine(DC, X, Ymin, Ymax);
	exit;
      end;
      YminPrev := Ymin;
      YmaxPrev := Ymax;
    end;
  end;
end;

procedure TCustomZoomDisplay.SetMonochromeBitmap(Value: Boolean);
begin
  if Value <> FMonochromeBitmap then
  begin
    if FBitmap <> nil then
    begin
      FBitmap.Monochrome := Value;
      FRestoreColors := True;
    end;
    FMonochromeBitmap := Value;
    InvalidateWorldWindow;
  end;
end;

procedure TCustomZoomDisplay.SetPaintMode(Value: TSignalPaintMode);
begin
  if Value <> FPaintMode then
  begin
    FPaintMode := Value;
    InvalidateWorldWindow;
  end;
end;

procedure TCustomZoomDisplay.DrawHLSample;
var
  OldPen: TPen;
begin
  if FHLSampleValid then
  if FHighlightedX <> -1 then
  with WaveformCanvas do
  begin
    OldPen := TPen.Create;
    try
      OldPen.Assign(Pen);
      try
        if (FBitmap <> nil) and FMonochromeBitmap then
          Pen.Mode := pmNot
        else
        begin
          Pen.Mode := pmNotXor;
          Pen.Color := FHighlightColor;
        end;
        Pen.Style := psSolid;
	MoveTo(FHighlightedX, 0);
        LineTo(FHighlightedX, FViewport.H+1);
      finally
        Pen := OldPen;
      end;
    finally
      OldPen.Free;
    end;
  end;
end;

procedure TCustomZoomDisplay.InvalidateHLSample;
var
  R: TRect;
begin
  if FHLSampleValid and (FHighlightedX <> -1) then
  begin
    if (FBitmap <> nil) and not FBitmapInvalid then DrawHLSample;
    R := HighlightedRect;
    InvalidateRect(Handle, @R, False);
  end;
end;

procedure TCustomZoomDisplay.HideHighlightedSample;
begin
  if FHLSampleValid then
  begin
    InvalidateHLSample;
    FHLSampleValid := False;
  end;
end;

function TCustomZoomDisplay.HighlightedRect: TRect;
begin
  Result.Left   := FHighlightedX + FViewport.X;
  Result.Right  := Result.Left+1;
  Result.Top    := FViewport.Y;
  Result.Bottom := FViewport.Y+FViewport.H+1;
end;

procedure TCustomZoomDisplay.SetHighlightColor(Value: TColor);
begin
  if FHighlightColor <> Value then
  begin
    FHighlightColor := Value;
    InvalidateHLSample;
  end;
end;

procedure TCustomZoomDisplay.SetHighlightedSample(Value: Integer);
begin
  if Value <> FHLSample then
  begin
    InvalidateHLSample;
    FHLSample := Value;
    FHLSampleValid := True;
    UpdateHighlightedX;
    InvalidateHLSample;
  end;
end;

procedure TCustomZoomDisplay.UpdateHighlightedX;
begin
  with Window do
  if (FHLSample < Left) or (FHLSample > Right)
    then FHighlightedX := -1
    else FHighlightedX := Round(DeviceX(FHLSample));
end;

procedure TCustomZoomDisplay.ScaleDrawGrid(
  Sender: TObject;
  ACanvas: TCanvas;
  Points: array of TPoint);
var
  OldPen: TPen;
  i: Integer;
begin
  if not TDisplayScale(Sender).FShowGrid then Exit;
  OldPen := TPen.Create;
  try
    OldPen.Assign(ACanvas.Pen);
    ACanvas.Pen := GridPen;
    try
      if GridPen.Color = clDefault then
	ACanvas.Pen.Color := DecentContrast(Color);
      if Sender = HorzScale then
        for i := Low(Points) to High(Points) do
        with Points[i] do
        begin
          ACanvas.MoveTo(X, 0);
          ACanvas.LineTo(X, Y);
        end
      else
        for i := Low(Points) to High(Points) do
        with Points[i] do
        begin
          ACanvas.MoveTo(FViewPort.W, Y);
          ACanvas.LineTo(0, Y);
	end;
    finally
      ACanvas.Pen := OldPen;
    end;
  finally
    OldPen.Free;
  end;
end;

procedure TCustomZoomDisplay.SetInfo(Value: TComponentInfo);
begin
  { dummy method }
end;

class function TCustomZoomDisplay.Version: Integer;
begin
  Result := $03020500;
end;

type
  TSignalDisplayEditor = class(TDefaultEditor)
  public
    procedure Edit; override;
  end;

procedure TSignalDisplayEditor.Edit;
begin
  TCustomZoomDisplay(Component).Info.Display;
end;

const
{$IFDEF Bilch}
  PalettePage = 'Homebrew';
{$ELSE}
  PalettePage = '3rdParty';
{$ENDIF}
procedure Register;
begin
  RegisterComponents(PalettePage, [TZoomDisplay]);
  RegisterComponentEditor(TCustomZoomDisplay, TSignalDisplayEditor);
end;
{$IFDEF Profile}
var
  ZProfiler: TZProfiler;
{$ENDIF}
initialization
  Screen.Cursors[crCrossHairs] := LoadCursor(HInstance, 'CROSSHAIRS');
  Screen.Cursors[crHand] := LoadCursor(HInstance, 'OPENHAND');
  Screen.Cursors[crGrip] := LoadCursor(HInstance, 'GRIP');
{$IFDEF Profile}
  ZProfiler := TZProfiler.Create(Application.MainForm);
finalization
  ZProfiler.Free;
{$ENDIF}
end.

